#!/usr/bin/env python
"""
Updated for python3 and scipy 1.0.0
"""
# u = TVRegDiff( data, iter, alph, u0, scale, ep, dx, plotflag, diagflag );
# 
# Rick Chartrand (rickc@lanl.gov), Apr. 10, 2011
# Please cite Rick Chartrand, "Numerical differentiation of noisy,
# nonsmooth data," ISRN Applied Mathematics, Vol. 2011, Article ID 164564, 
# 2011. 
#
# Inputs:  (First three required; omitting the final N parameters for N < 7
#           or passing in [] results in default values being used.) 
#       data        Vector of data to be differentiated.
#
#       iter        Number of iterations to run the main loop.  A stopping
#                   condition based on the norm of the gradient vector g
#                   below would be an easy modification.  No default value.
#
#       alph        Regularization parameter.  This is the main parameter
#                   to fiddle with.  Start by varying by orders of
#                   magnitude until reasonable results are obtained.  A
#                   value to the nearest power of 10 is usally adequate.
#                   No default value.  Higher values increase
#                   regularization strenght and improve conditioning.
#
#       u0          Initialization of the iteration.  Default value is the
#                   naive derivative (without scaling), of appropriate
#                   length (this being different for the two methods).
#                   Although the solution is theoretically independent of
#                   the intialization, a poor choice can exacerbate
#                   conditioning issues when the linear system is solved.
#
#       scale       'large' or 'small' (case insensitive).  Default is
#                   'small'.  'small' has somewhat better boundary
#                   behavior, but becomes unwieldly for data larger than
#                   1000 entries or so.  'large' has simpler numerics but
#                   is more efficient for large-scale problems.  'large' is
#                   more readily modified for higher-order derivatives,
#                   since the implicit differentiation matrix is square.
#
#       ep          Parameter for avoiding division by zero.  Default value
#                   is 1e-6.  Results should not be very sensitive to the
#                   value.  Larger values improve conditioning and
#                   therefore speed, while smaller values give more
#                   accurate results with sharper jumps.
#
#       dx          Grid spacing, used in the definition of the derivative
#                   operators.  Default is the reciprocal of the data size.
#
#       plotflag    Flag whether to display plot at each iteration.
#                   Default is 1 (yes).  Useful, but adds significant
#                   running time.
#
#       diagflag    Flag whether to display diagnostics at each
#                   iteration.  Default is 1 (yes).  Useful for diagnosing
#                   preconditioning problems.  When tolerance is not met,
#                   an early iterate being best is more worrying than a
#                   large relative residual.
#                   
# Output:
#
#       u           Estimate of the regularized derivative of data.  Due to
#                   different grid assumptions, length( u ) = 
#                   length( data ) + 1 if scale = 'small', otherwise
#                   length( u ) = length( data ).

## Copyright notice:
# Copyright 2010. Los Alamos National Security, LLC. This material
# was produced under U.S. Government contract DE-AC52-06NA25396 for
# Los Alamos National Laboratory, which is operated by Los Alamos
# National Security, LLC, for the U.S. Department of Energy. The
# Government is granted for, itself and others acting on its
# behalf, a paid-up, nonexclusive, irrevocable worldwide license in
# this material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# (March 31, 2011) permission to assert copyright was obtained,
# subject to additional five-year worldwide renewals, the
# Government is granted for itself and others acting on its behalf
# a paid-up, nonexclusive, irrevocable worldwide license in this
# material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and
# to permit others to do so. NEITHER THE UNITED STATES NOR THE
# UNITED STATES DEPARTMENT OF ENERGY, NOR LOS ALAMOS NATIONAL
# SECURITY, LLC, NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY,
# EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL LIABILITY OR
# RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR USEFULNESS OF
# ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS DISCLOSED, OR
# REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY OWNED
# RIGHTS. 

## BSD License notice:
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met: 
# 
#      Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.  
#      Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution. 
#      Neither the name of Los Alamos National Security nor the names of its
#      contributors may be used to endorse or promote products
#      derived from this software without specific prior written
#      permission. 
#  
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE. 
#
#########################################################
#                                                       #
# Python translation by Simone Sturniolo                #
# Rutherford Appleton Laboratory, STFC, UK (2014)       #
# simonesturniolo@gmail.com                             #
#                                                       #
#########################################################

import numpy as np
from scipy import sparse
from scipy.sparse import linalg as splin


def tv_reg_diff(data, itern, alph, u0=None, scale='small', ep=1e-6, dx=None, diagflag=1):
    ## code starts here
    # Make sure we have a column vector
    data = np.array(data)
    if len(data.shape) != 1:
        print("Error - data is not a column vector")
        return
    # Get the data size.
    n = len(data)

    # Default checking. (u0 is done separately within each method.)
    if dx is None:
        dx = 1.0 / n

    # Different methods for small- and large-scale problems.
    if scale.lower() == 'small':

        # Construct differentiation matrix.
        c = np.ones(n + 1) / dx
        D = sparse.spdiags([-c, c], [0, 1], n, n + 1)

        DT = D.transpose()

        # Construct antidifferentiation operator and its adjoint.
        A = lambda x: (np.cumsum(x) - 0.5 * (x + x[0]))[1:] * dx
        AT = lambda w: (sum(w) * np.ones(n + 1) - np.transpose(
            np.concatenate(([sum(w) / 2.0], np.cumsum(w) - w / 2.0)))) * dx

        # Default initialization is naive derivative

        if u0 is None:
            u0 = np.concatenate(([0], np.diff(data), [0]))

        u = u0
        # Since Au( 0 ) = 0, we need to adjust.
        ofst = data[0]
        # Precompute.
        ATb = AT(ofst - data)  # input: size n

        # Main loop.
        for ii in range(1, itern + 1):
            # Diagonal matrix of weights, for linearizing E-L equation.
            Q = sparse.spdiags(1. / (np.sqrt((D * u) ** 2 + ep)), 0, n, n)
            # Linearized diffusion matrix, also approximation of Hessian.
            L = dx * DT * Q * D

            # Gradient of functional.
            g = AT(A(u)) + ATb + alph * L * u

            # Prepare to solve linear equation.
            tol = 1e-4
            maxit = 100
            # Simple preconditioner.
            P = alph * sparse.spdiags(L.diagonal() + 1, 0, n + 1, n + 1)

            linop = lambda v: (alph * L * v + AT(A(v)))
            linop = splin.LinearOperator((n + 1, n + 1), linop)

            if diagflag:
                [s, info_i] = sparse.linalg.cg(A=linop, b=g, x0=None, tol=tol, maxiter=maxit, M=P)
                print('iteration {0:4d}: relative change = {1:.3e}, gradient norm = {2:.3e}\n'.format(ii,
                                                                                                      np.linalg.norm(s[
                                                                                                                         0]) / np.linalg.norm(
                                                                                                          u),
                                                                                                      np.linalg.norm(
                                                                                                          g)))
                if (info_i > 0):
                    print("WARNING - convergence to tolerance not achieved!")
                elif (info_i < 0):
                    print("WARNING - illegal input or breakdown")
            else:
                [s, info_i] = sparse.linalg.cg(A=linop, b=g, x0=None, tol=tol, maxiter=maxit, M=P)
            # Update solution.
            u = u - s;
            # Display plot.

    elif scale.lower() == 'large':
        # Construct antidifferentiation operator and its adjoint.
        A = lambda v: np.cumsum(v)
        AT = lambda w: (sum(w) * np.ones(len(w)) - np.transpose(np.concatenate(([0.0], np.cumsum(w[:-1])))))
        # Construct differentiation matrix.
        c = np.ones(n)
        D = sparse.spdiags([-c, c], [0, 1], n, n) / dx
        mask = np.ones((n, n))
        mask[-1, -1] = 0.0
        D = sparse.dia_matrix(D.multiply(mask))
        DT = D.transpose()
        # Since Au( 0 ) = 0, we need to adjust.
        data = data - data[0]
        # Default initialization is naive derivative.
        if u0 is None:
            u0 = np.concatenate(([0], np.diff(data)))
        u = u0
        # Precompute.
        ATd = AT(data)

        # Main loop.
        for ii in range(1, itern + 1):
            # Diagonal matrix of weights, for linearizing E-L equation.
            Q = sparse.spdiags(1. / np.sqrt((D * u) ** 2.0 + ep), 0, n, n)
            # Linearized diffusion matrix, also approximation of Hessian.
            L = DT * Q * D
            # Gradient of functional.
            g = AT(A(u)) - ATd
            g += alph * L * u
            # Build preconditioner.
            c = np.cumsum(range(n, 0, -1))
            B = alph * L + sparse.spdiags(c[::-1], 0, n, n)
            # droptol = 1.0e-2
            R = sparse.dia_matrix(np.linalg.cholesky(B.todense()))
            # Prepare to solve linear equation.
            tol = 1.0e-4
            maxit = 100

            linop = lambda v: (alph * L * v + AT(A(v)))
            linop = splin.LinearOperator((n, n), linop)

            if diagflag:
                [s, info_i] = sparse.linalg.cg(A=linop, b=-g, x0=None, tol=tol, maxiter=maxit,
                                               M=np.dot(R.transpose(), R))
                print('iteration {0:4d}: relative change = {1:.3e}, gradient norm = {2:.3e}\n'.format(ii,
                                                                                                      np.linalg.norm(s[
                                                                                                                         0]) / np.linalg.norm(
                                                                                                          u),
                                                                                                      np.linalg.norm(
                                                                                                          g)))
                if (info_i > 0):
                    print("WARNING - convergence to tolerance not achieved!")
                elif (info_i < 0):
                    print("WARNING - illegal input or breakdown")

            else:
                [s, info_i] = sparse.linalg.cg(A=linop, b=-g, x0=None, tol=tol, maxiter=maxit,
                                               M=np.dot(R.transpose(), R))
            # Update current solution
            u = u + s

        u = u / dx

    return u


# Small testing script

if __name__ == "__main__":
    import matplotlib.pyplot as plt

    x0 = np.arange(0, 2.0 * np.pi, 0.05)

    testf = [np.sin(x) for x in x0]

    testf = testf + np.random.normal(0.0, 0.04, x0.shape)
    deriv_sm = tv_reg_diff(testf, 100, 5e-2, dx=0.05, ep=1e-1, scale='small', diagflag=0)
    deriv_lrg = tv_reg_diff(testf, 100, 1e-1, dx=0.05, ep=1e-2, scale='large', diagflag=0)

    plt.plot(testf)
    plt.plot(deriv_sm, c='r')
    plt.plot(deriv_lrg, c='g')
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb 24 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import pandas as pd
import pickle
import json
from copy import deepcopy
from uuid import uuid4, UUID
from typing import Tuple, List, Dict, Union, Optional
from itertools import chain
import os
import traceback
from warnings import warn
from configparser import RawConfigParser
from ..common.utils import HdfTools, draw_graph
from ..common import get_proj_config
from ..common.configuration import get_environment
from tqdm import tqdm
from functools import lru_cache
import logging


logger = logging.getLogger()


class _HistoryTraceExceptions(Exception):
    def __init__(self, msg):
        assert isinstance(msg, str)
        self.msg = msg

    def __str__(self):
        return str(self.__doc__) + '\n' + self.msg


class DataBlockNotFound(_HistoryTraceExceptions):
    """Requested data block not found"""


class DataBlockAlreadyExists(_HistoryTraceExceptions):
    """Data block already exists in HistoryTrace."""


class OperationNotFound(_HistoryTraceExceptions):
    """Requested operation not found in data block."""


class HistoryTrace:
    """
    Structure of a history trace:

    A dict with keys that are the block_ids. Each dict value is a list of operation_dicts.
    Each operation_dict has a single key which is the name of the operation and the value of that key is the operation parameters.

        {block_id_1: [\n
                        {operation_1:\n
                            {\n
                             param_1: a,\n
                             param_2: b,\n
                             param_n, z\n
                             }\n
                         },\n
\n
                        {operation_2:\n
                            {\n
                             param_1: a,\n
                             param_n, z\n
                             }\n
                         },\n
                         ...\n
                        {operation_n:\n
                            {\n
                             param_n: x\n
                             }\n
                         }\n
                     ]\n
         block_id_2: <list of operation dicts>,\n
         ...\n
         block_id_n: <list of operation dicts>\n
         }\n

    **The main dict illustrated above should never be worked with directly.**\n
    **You must use the helper methods of this class to query or add information**

    """
    def __init__(self, history: Dict[Union[UUID, str], List[Dict]] = None, data_blocks: List[Union[UUID, str]] = None):
        """
        :param history:     Dict containing a data block UUIDs as keys. The values are a list of dicts containing operation parameters.
        :param data_blocks: List of data block UUIDs

        :ivar _history:     The dict of the actual data, as illustrated above. Should not be accessed directly. Use the :py:attr:`~history` property or call `get_all_data_blocks_history()`.
        :ivar _data_blocks: List of all data blocks. Should not be accessed directly, use the :py:attr:`~data_blocks` property instead.
        """
        self._history = None
        self._data_blocks = None

        if None in [history, data_blocks]:
            self.data_blocks = []
            self.history = dict()
        else:
            for i in range(len(data_blocks)):
                data_blocks[i] = self._to_uuid(data_blocks[i])
            self.data_blocks = data_blocks
            for k in list(history.keys()):
                history[self._to_uuid(k)] = history.pop(k)
            self.history = history

    @property
    def data_blocks(self) -> list:
        """List of UUIDs that allow you to pin down the history of specific rows of the dataframe to their history
        as stored in the history trace data structure (self.history)"""
        return self._data_blocks

    @data_blocks.setter
    def data_blocks(self, dbl: list):
        self._data_blocks = dbl

    @property
    def history(self) -> dict:
        """The analysis log that is stored in the structure outlined in the doc string"""
        return self._history

    @history.setter
    def history(self, h):
        self._history = h

    def create_data_block(self, dataframe: pd.DataFrame) -> Tuple[pd.DataFrame, UUID]:
        """
        Creates a new UUID, assigns it to the input dataframe by setting the UUID in the _BLOCK_ column

        :param dataframe: Assigns a block ID to this entire DataFrame.
        """
        block_id = uuid4()
        self._add_data_block(block_id)
        dataframe['_BLOCK_'] = str(block_id)
        return dataframe, block_id

    def _add_data_block(self, data_block_id: UUID):
        """
        Adds new datablock UUID to the list of datablocks in this instance.
        Throws exception if UUID already exists.
        """
        assert isinstance(data_block_id, UUID)
        if data_block_id in self.data_blocks:
            raise DataBlockAlreadyExists(str(data_block_id))
        else:
            self.data_blocks.append(data_block_id)

        self.history.update({data_block_id: []})

    def add_operation(self, data_block_id: Union[UUID, str], operation: str, parameters: dict):
        """
        Add a single operation, that is usually performed by a node, to the history trace.
        Added to all or specific datablock(s), depending on which datablock(s) the node performed the operation on

        :param data_block_id: data_block_id to log the operation on to. either a UUID or 'all' to append the operation to all data blocks
        :param operation: name of the operation, usually the same as the name of the node in all lowercase
        :param parameters: operation parameters.
        """
        assert isinstance(operation, str)
        assert isinstance(parameters, dict)

        if isinstance(data_block_id, str):
            if data_block_id == 'all':
                _ids = self.data_blocks
            else:
                try:
                    _ids = [self._to_uuid(data_block_id)]
                except ValueError:
                    raise ValueError("data_block_id must be a UUID, str representation of a UUID, or 'all' ")

        else:
            _ids = [data_block_id]

        if not all(u in self.data_blocks for u in _ids):
            raise DataBlockNotFound()
        for _id in _ids:
            self.history[_id].append({operation: parameters})

    def get_data_block_history(self, data_block_id: Union[str, UUID], copy: bool = False) -> List[dict]:
        """
        Get the full history trace of a single data block.

        Use copy=False if you want to modify the history trace of the data block.

        :param data_block_id: data block ID
        :type data_block_id: Union[str, UUID]

        :param copy: If true, returns a deepcopy
        :type copy: bool

        :return: data block history
        :rtype: List[dict]
        """

        data_block_id = self._to_uuid(data_block_id)

        if data_block_id not in self.data_blocks:
            raise DataBlockNotFound(str(data_block_id))

        if copy:
            return deepcopy(self.history[data_block_id])

        else:
            return self.history[data_block_id]

    def get_all_data_blocks_history(self) -> dict:
        """Returns history trace of all datablocks"""
        h = {}

        for block_id in self.data_blocks:
            h.update({str(block_id): self.get_data_block_history(block_id)})

        return h.copy()

    def get_operations_list(self, data_block_id: Union[UUID, str]) -> list:
        """
        Returns just a simple list of operations in the order that they were performed on the given datablock.
        To get the operations along with their parameters call get_data_block_history()
        """
        data_block_id = self._to_uuid(data_block_id)

        l = [next(iter(d)) for d in self.get_data_block_history(data_block_id)]
        return l

    def get_operation_params(self, data_block_id: Union[UUID, str], operation: str) -> dict:
        """Get the parameters dict for a specific operation that was performed on a specific data block"""
        # if isinstance(data_block_id, str):
        #     data_block_id = UUID(data_block_id)
        data_block_id = self._to_uuid(data_block_id)

        try:
            l = self.get_data_block_history(data_block_id)
            params = next(d for ix, d in enumerate(reversed(l)) if operation in d)[operation]
        except StopIteration:
            raise OperationNotFound('Data block: ' + str(data_block_id) + ', Operation: ' + operation)

        return params.copy()

    def check_operation_exists(self, data_block_id: UUID, operation: str) -> bool:
        """Check if a specific operation was performed on a specific datablock"""
        data_block_id = self._to_uuid(data_block_id)
        try:
            self.get_operation_params(data_block_id, operation)
        except OperationNotFound:
            return False
        else:
            return True

    @staticmethod
    def _to_uuid(u: Union[str, UUID]) -> UUID:
        """
        If argument 'u' is type <str> that can be formatted as a UUID, return it as UUID type.
        If argument 'u' is a UUID, just return it.
        """
        if isinstance(u, UUID):
            return u
        elif isinstance(u, str):
            return UUID(u)
        else:
            raise TypeError('Must pass str or UUID')

    def to_dict(self) -> dict:
        """
        Package the HistoryTrace instance as a dict. Converts all UUIDs to <str> representation for JSON compatibility.
        """

        dbs_str = [str(db) for db in self.data_blocks]
        hist_str = self.get_all_data_blocks_history()

        return {'history': hist_str, 'data_blocks': dbs_str}

    @staticmethod
    def from_dict(d: dict) -> dict:
        """
        Format a dict stored using HistoryTrace.to_dict so that it can be used to create a HistoryTrace instance.
        Converts all the <str> representations of UUID back to :class:`<uuid.UUID>` types.

        :param d: dict containing appropriate 'history' and 'datablocks' keys. Must be packaged by HistoryTrace.to_dict()
        :return: dict formatted so that it can be used to instantiate a HistoryTrace instance recapitulating the HistoryTrace it was packaged from.
        """

        hist = {HistoryTrace._to_uuid(k): v for k, v in d['history'].items()}
        dbs = [HistoryTrace._to_uuid(u) for u in d['data_blocks']]

        return {'history': hist, 'data_blocks': dbs}

    def to_json(self, path: str):
        """
        Save HistoryTrace to a JSON file.

        :param path: file path, usually ends with .json
        """

        json.dump(self.to_dict(), open(path, 'w'))

    @classmethod
    def from_json(cls, path: str):
        """
        Instantiate HistoryTrace from JSON file (that was saved using HistoryTrace.to_json)

        :param path: file path, usually ends with .json
        """

        j = json.load(open(path, 'r'))
        j = HistoryTrace.from_dict(j)

        return cls(history=j['history'], data_blocks=['data_blocks'])

    def to_pickle(self, path: str):
        """
        Dump this instance to a pickle

        :param path: file path
        """
        pickle.dump(self.to_dict(), open(path, 'wb'))

    @classmethod
    def from_pickle(cls, path: str):
        """
        Load HistoryTrace that was pickled

        :param path: file path
        """
        p = pickle.load(open(path, 'r'))
        p = HistoryTrace.from_dict(p)
        return cls(history=p['history'], data_blocks=p['data_blocks'])

    @classmethod
    def merge(cls, history_traces: list):
        """
        Merge a list of HistoryTrace instances into one HistoryTrace instance.
        Useful when merging Transmission objects.

        :param history_traces: list of HistoryTrace instances
        """

        assert all(isinstance(h, HistoryTrace) for h in history_traces)
        data_blocks_l2_list = [h.data_blocks for h in history_traces]
        data_blocks = list(chain.from_iterable(data_blocks_l2_list))

        history = dict()
        for h in history_traces:
            d = h.history
            history.update(d)

        return cls(history=history, data_blocks=data_blocks)

    def draw_graph(self, data_block_id: Union[str, UUID], **kwargs) -> str:
        """
        Draw graph of a data block. kwargs are passed to `mesmerize.common.utils.draw_graph`

        :param data_block_id: data block ID from which to get history to draw in a graph
        :type data_block_id: Union[str, UUID]

        :return: file path to the graph pdf file
        :rtype: str
        """
        db_history = self.get_data_block_history(data_block_id, copy=True)

        cleaned = HistoryTrace.clean_history_trace(db_history)

        return draw_graph(cleaned, **kwargs)

    @staticmethod
    def clean_history_trace(db_history: list) -> list:
        """
        Cleans up excessive data such as frequencies linspaces and linkage matrices so the graph is viewable.

        :param db_history: data block history
        :return: data block history with excessive params removed
        """

        for op in db_history:
            o = list(op.keys())[0]
            params = op[o]
            op[o] = HistoryTrace._cleanup_params(o, params)

        return db_history

    @staticmethod
    def _cleanup_params(operation, params):
        log = params.copy()
        if operation == 'rfft':
            log.pop('frequencies')
        elif operation == 'fcluster':
            log.pop('linkage_matrix')

        return log


class BaseTransmission:
    def __init__(self, df: pd.DataFrame, history_trace: HistoryTrace, proj_path: str = None, last_output: str = None,
                 last_unit: str = None, ROI_DEFS: list = None, STIM_DEFS: list = None, CUSTOM_COLUMNS: list = None,
                 plot_state: dict = None):
        """
        Base class for common Transmission functions

        :param  df:             Transmission dataframe
        :param  history_trace:  HistoryTrace object, keeps track of the nodes & node parameters
                                the transmission has been processed through
        :param  proj_path:      Project path, necessary for the datapoint tracer
        :param  last_output:    Last data column that was appended via a node's operation
        :param  last_unit:      Current units of the data. Refers to the units of column in last_output
        :param plot_state:      State of a plot, such as data and label columns. Used when saving interactive plots.

        :type df:               pd.DataFrame
        :type history_trace:    HistoryTrace
        :type proj_path:        str
        :type last_output:      str
        :type last_unit:        str
        :type plot_state:       dict

        :ivar df:               DataFrame instance
        :ivar history_trace:    :class:`HistoryTrace instance <mesmerize.analysis.data_types.HistoryTrace>`
        :ivar last_output:      Name of the DataFrame column that contains data from the most recent node
        :ivar last_unit:        The data units for the data in the column of 'last_output'
        :ivar plot_state:       State of a plot, containing user entered plot parameters. Used for storing interactive plot states.
        """

        self.df = df

        if isinstance(history_trace, HistoryTrace):
            self.history_trace = history_trace
        elif isinstance(history_trace, dict):
            self.history_trace = HistoryTrace(**history_trace)

        self._proj_path = None
        if proj_path is not None:
            self.set_proj_path(proj_path)

        self.last_output = last_output
        self.last_unit = last_unit

        if ROI_DEFS is None:
            self.ROI_DEFS = []
        else:
            assert isinstance(ROI_DEFS, list)
            self.ROI_DEFS = ROI_DEFS

        if STIM_DEFS is None:
            self.STIM_DEFS = []
        else:
            assert isinstance(STIM_DEFS, list)
            self.STIM_DEFS = STIM_DEFS

        if CUSTOM_COLUMNS is None:
            self.CUSTOM_COLUMNS = []
        else:
            assert isinstance(CUSTOM_COLUMNS, list)
            self.CUSTOM_COLUMNS = CUSTOM_COLUMNS

        self.plot_state = plot_state  #: If used in a plot, dict containing information about the plot state

    def to_dict(self) -> dict:
        """
        Package Transmission as a dict, useful for saving to hdf5 or pickle
        """
        d = {'df':              self.df,
             'history_trace':   self.history_trace.to_dict(),
             'last_output':     self.last_output,
             'last_unit':       self.last_unit,
             'plot_state':      self.plot_state
             }

        return d

    def to_hdf5(self, path: str):
        """
        Save as an hdf5 file. Uses pytables to save the DataFrame, serielizes the HistoryTrace using JSON.
        See :class:`HdfTools <mesmerize.common.utils.HdfTools>`

        :param path: file path, usually ends in .trn
        """
        d = self.to_dict()
        df = d.pop('df')
        HdfTools.save_dataframe(path=path, dataframe=df, metadata=d, metadata_method='json')

    @classmethod
    def from_hdf5(cls, path: str):
        """
        Create Transmission from an hdf5 file. See :class:`HdfTools <mesmerize.common.utils.HdfTools>` for information on the file structure.

        :param path: file path, usually ends in .trn (.ptrn for plots)
        """
        df, meta = HdfTools.load_dataframe(path)
        return cls(df, **meta)

    @classmethod
    def from_pickle(cls, path):
        """
        Load Transmission from a pickle.

        :param path: file path, usually ends in .trn
        """
        p = pickle.load(open(path, 'rb'))
        return cls(**p)

    def to_pickle(self, path: str):
        """
        Save Transmission as a pickle.  Not recommended for sharing data, use :func:`~to_hdf5`

        :param path: file path, usually ends in .trn
        """
        pickle.dump(self.to_dict(), open(path, 'wb'), protocol=4)

    def copy(self):
        return deepcopy(self)

    @staticmethod
    def empty_df(transmission, addCols: list = None) -> pd.DataFrame:
        """
        Just a helper method to return an empty DataFrame with the same columns

        :param transmission: Transmission object that forms the basis
        :param addCols: list of columns to add

        :return: The input transmission with an empty dataframe containing the same columns and any additional columns that were passed

        """
        if addCols is None:
            addCols = []

        c = list(transmission.df.columns) + addCols
        e_df = pd.DataFrame(columns=c)
        return e_df
        # return cls(e_df, transmission.history_trace, **transmission.kwargs)

    def get_proj_path(self) -> str:
        """
        Get the project root dir associated to this Transmission.

        :return: Root directory of the project
        """
        if self._proj_path is None:
            raise ValueError('No project path set')
        return self._proj_path

    def set_proj_path(self, path: str):
        """
        Set the project root dir for this transmission.

        Used for finding associated project files, for example the Datapoint Tracer uses it to find max and std projections of image sequences.

        :param path: Root directory of the project
        """

        path = os.path.abspath(path)

        if not os.path.isdir(path + '/images'):
            raise NotADirectoryError('images directory not found')
        if not os.path.isfile(path + '/config.cfg'):
            raise FileNotFoundError('Project config not found')

        self._proj_path = path

        self.set_proj_config()

    def set_proj_config(self):
        """
        Sets some project config related attributes from the project's config file.
        """
        proj_path = self.get_proj_path()
        if proj_path is None:
            raise ValueError('Project path must be set before setting project configuration')

        config_path = proj_path + '/config.cfg'

        proj_config = RawConfigParser(allow_no_value=True)
        proj_config.optionxform = str
        proj_config.read(config_path)

        self.ROI_DEFS = proj_config.options('ROI_DEFS')
        self.STIM_DEFS = proj_config.options('STIM_DEFS')
        self.CUSTOM_COLUMNS = proj_config.options('CUSTOM_COLUMNS')


class Transmission(BaseTransmission):
    """The transmission object used throughout the flowchart"""
    @classmethod
    def from_proj(cls, proj_path: str, dataframe: pd.DataFrame, sub_dataframe_name: str = 'root',
                  dataframe_filter_history: dict = None):
        """
        :param proj_path: root directory of the project
        :param dataframe: Chosen Child DataFrame from the Mesmerize Project
        :param sub_dataframe_name: Name of the sub DataFrame to load
        :param dataframe_filter_history: Filter history of the child dataframe

        """
        df = dataframe.copy()
        logger.info('Collecting image metadata')
        tqdm().pandas()
        df[['meta', 'stim_maps']] = df.progress_apply(lambda r: Transmission._load_files(proj_path, r), axis=1)

        Transmission._load_imginfo.cache_clear()

        try:
            logger.info('Collecting curve data')
            df['_RAW_CURVE'] = df['ROI_State'].progress_apply(
                    lambda r: r['curve_data'][1]
                )
        except:
            raise IndexError("Curve data missing from one of your samples.\n"
                             "See the progress bar to get the row index of "
                             "the project dataframe where the curve data are missing")

        try:
            logger.info('Collecting curve data')
            df['_SPIKES'] = df['ROI_State'].progress_apply(
                lambda r: r['spike_data'][1] if (r['spike_data'] is not None) else None
            )

        except KeyError:
            warn('spikes or data not found, probably is probably from Mesmerize version < 0.2')

        try:
            logger.info('Collecting dfof data')
            df['_DFOF'] = df['ROI_State'].progress_apply(
                lambda r: r['dfof_data'][1] if (r['dfof_data'] is not None) else None
            )
        except KeyError:
            warn('dfof data not found, probably is probably from Mesmerize version < 0.2')

        df.sort_values(by=['SampleID'], inplace=True)
        df = df.reset_index(drop=True)

        h = HistoryTrace()
        df, block_id = h.create_data_block(df)

        params = {'sub_dataframe_name': sub_dataframe_name, 'dataframe_filter_history': dataframe_filter_history}
        h.add_operation(data_block_id=block_id, operation='spawn_transmission', parameters=params)

        proj_config = get_proj_config(proj_path)

        try:
            roi_type_defs = proj_config.options('ROI_DEFS')
            stim_type_defs = proj_config.options('STIM_DEFS')
            custom_columns = proj_config.options('CUSTOM_COLUMNS')
        except:
            raise ValueError('Could not read project configuration when creating Transmission'
                             '\n' + traceback.format_exc())

        return cls(df, proj_path=proj_path, history_trace=h, last_output=None, last_unit='time',
                   ROI_DEFS=roi_type_defs, STIM_DEFS=stim_type_defs, CUSTOM_COLUMNS=custom_columns)

    @staticmethod
    @lru_cache(maxsize=get_environment().LRU_CACHE_MAXSIZE)
    def _load_imginfo(path: str) -> dict:
        return pickle.load(open(path, 'rb'))

    @staticmethod
    def _load_files(proj_path: str, row: pd.Series) -> pd.Series:
        """Loads npz of curve data and pickle files containing metadata using the paths specified in each row of the
        chosen sub-dataframe of the project"""

        pik_path = os.path.join(proj_path, row['ImgInfoPath'])
        pik = Transmission._load_imginfo(pik_path)
        meta = pik['meta']
        stim_maps = pik['stim_maps']

        return pd.Series(
            {
                'meta': meta,
                'stim_maps': [[stim_maps]]
            }
        )

    def get_data_block_dataframe(self, data_block_id: Union[UUID, str]) -> pd.DataFrame:
        """
        Get the DataFrame rows corresponding to a single data block.

        :param data_block_id: data block uuid
        :return: DataFrame with rows from the specified data block
        """
        if isinstance(data_block_id, UUID):
            data_block_id = str(data_block_id)
        assert isinstance(data_block_id, str)

        if UUID(data_block_id) not in self.history_trace.data_blocks:
            raise DataBlockNotFound(data_block_id)

        return self.df[self.df._BLOCK_ == data_block_id]

    @classmethod
    def merge(cls, transmissions: list):
        """
        Merges a list of Transmissions into one transmission. A single DataFrame is created by simple concatenation.
        HistoryTrace objects are also merged using HistoryTrace.merge.

        :param transmissions: A list containing Transmission objects to merge
        :return: Merged transmission
        """
        proj_path_list = [os.path.abspath(t.get_proj_path()) for t in transmissions]
        if len(set(proj_path_list)) > 1:
            raise ValueError('You cannot merge transmissions from different projects. '
                             'You have tried to merge transmissions from the following projects: '
                             + '\n'.join(set(proj_path_list)))
        else:
            proj_path = proj_path_list[0]
            roi_defs = transmissions[0].ROI_DEFS
            stim_defs = transmissions[0].STIM_DEFS
            custom_columns = transmissions[0].CUSTOM_COLUMNS

        units = set([t.last_unit for t in transmissions])

        if len(units) > 1:
            raise ValueError('Cannot merge transmissions of differing data units. The inputs have the following '
                             'units in their last output data column: \n' + str(units))

        dfs = [t.df for t in transmissions]
        df = pd.concat(dfs, copy=True)

        h = [t.history_trace for t in transmissions]
        h = HistoryTrace.merge(h)

        return cls(df, proj_path=proj_path, history_trace=h,
                   ROI_DEFS=roi_defs, STIM_DEFS=stim_defs, CUSTOM_COLUMNS=custom_columns)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 16 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""


# from common.window_manager import WindowManager
from .. import pyqtgraphCore
from ..viewer import ViewerWindow
from ..common import get_window_manager
# from viewer.core.viewer_work_environment import ViewerWorkEnv
# from viewer.modules import tiff_io
from ..analysis.flowchart import Window as FlowchartWindow
# from analysis.stats_gui import StatsWindow
from ..project_manager import ProjectBrowserWindow
# import json
from uuid import UUID


# def window_manager():
#     get_window_manager().initalize()
#
#
# def main():
#     w = welcome_window.MainWindow()
#     get_window_manager().welcome_window = w
#     w.show()
#

# def background_batch(run_batch: list):
#     get_window_manager().get_batch_manager(run_batch)


def project_browser():
    project_browser_window = ProjectBrowserWindow()
    get_window_manager().project_browser = project_browser_window
    num_columns = len(project_browser_window.project_browser.tabs['root'].columns)
    project_browser_window.resize(min(1920, num_columns * 240), 600)
    return project_browser_window


# def load_child_dataframes_gui():
#     configuration.window_manager.project_browsers[0].reload_all_tabs()


def viewer(file: str = None, sample_id: str = None, uuid: UUID = None):
    pyqtgraphCore.setConfigOptions(imageAxisOrder='row-major')
    w = ViewerWindow()

    w.resize(1460, 950)
    # TODO: INTEGRATE VIEWER initiation INTO VIEWERWINDOW __init__
    viewer_widget = pyqtgraphCore.ImageView(parent=w)
    viewer_widget.setPredefinedGradient('flame')

    w.viewer_reference = viewer_widget
    w.setCentralWidget(viewer_widget)

    # w = get_window_manager().get_new_viewer_window()

    return w

    # if file is not None:
    #     if file.endswith('.tiff') or file.endswith('.tif'):
    #         viewerWindow.run_module(tiff_io.ModuleGUI)
    #         tm = viewerWindow.running_modules[-1]
    #         assert isinstance(tm, tiff_io.ModuleGUI)
    #         tm.ui.labelFileTiff.setText(file)
    #     elif file.endswith('.mes'):
    #         pass
    #     elif file.endswith('.vwe'):
    #         viewer_widget.workEnv = ViewerWorkEnv.from_pickle(pickle_file_path=file)
    #     else:
    #         raise ValueError('File extension not supported')
    # elif sample_id is not None:
    #     pass
    # elif uuid is not None:
    #     pass


def flowchart(file=None, parent=None) -> FlowchartWindow:
    w = FlowchartWindow(parent=parent, filename=file)

    return w


# def plots(file=None):
#     configuration.window_manager.plots.append(StatsWindow())
#     configuration.window_manager.plots[-1].show()
#
#     if file is None:
#         return
#
#     elif file.endswith('.strn'):
#         pass
#
#     elif file.endswith('.gtrn'):
#         pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri March 2 2017

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import sys

sys.setrecursionlimit(10000)
from PyQt5 import QtWidgets
from PyQt5.QtGui import QCursor
from ..pyqtgraphCore.flowchart import Flowchart
import numpy as np
import pandas as pd
from .pytemplates import mainwindow_pytemplate as uiWin
from ..pyqtgraphCore.console import ConsoleWidget
from ..common import configuration
import os


class Window(QtWidgets.QMainWindow, uiWin.Ui_MainWindow):
    def __init__(self, parent=None, filename: str = None):
        super().__init__()
        self.setupUi(self)

        self.dockConsole.hide()

        self.fc = Flowchart(terminals={'dataIn': {'io': 'in'}, 'dataOut': {'io': 'out'}})

        self.fc_widget = self.fc.widget()
        self.fc.setParent(self)

        ns = {'np': np,
              'pd': pd,
              'get_nodes': self.fc.nodes,
              'this': self
              }

        txt = "Namespaces:\n" \
              "pandas as 'pd'\n" \
              "numpy as 'np'\n" \
              "self as 'this'\n" \
              "'get_nodes()' to get a dict of all nodes\n"

        cmd_history_path = os.path.join(configuration.console_history_path, 'flowchart.pik')

        self.dockConsole.setWidget(ConsoleWidget(namespace=ns, text=txt,
                                                 historyFile=cmd_history_path))
        # self.resizeDocks([self.dockConsole], [235], QtCore.Qt.Vertical)

        self.dockFcWidget.setWidget(self.fc_widget)

        self.setCentralWidget(self.fc_widget.chartWidget)

        if filename is not None:
            self.fc.loadFile(filename)

        self.setAcceptDrops(True)

    def dragEnterEvent(self, QDragEnterEvent):
        if QDragEnterEvent.mimeData().hasUrls():
            QDragEnterEvent.accept()
        else:
            QDragEnterEvent.ignore()

    def dropEvent(self, ev):
        files = ev.mimeData().urls()
        if len(files) == 1:
            file = files[0].path()
        else:
            return

        if file.endswith('.trn') or file.endswith('.ptrn'):
            fname = os.path.splitext(os.path.basename(file))[0]
            from_global = self.fc_widget.chartWidget.view.mapFromGlobal(QCursor().pos())
            pos = self.fc_widget.chartWidget.view.mapToScene(from_global)
            self.fc.createNode('LoadFile', name=fname, pos=pos)
            self.fc.nodes()[fname].load_file(file)



from setuptools import setup, find_packages
from pathlib import Path


entry_points = {'console_scripts': ['mesmerize=mesmerize.__main__:main']}

install_requires = \
    [
        "cycler==0.10.0",
        "dask>=1.1.1",
        "future>=0.18.2",
        "h5py~=2.10.0",
        "ipykernel>=4.10",
        "ipython>=7.15.0",
        "ipyparallel>=6.3",
        "MarkupSafe>=1.1.1",
        "matplotlib<=3.2.1",
        "opencv-contrib-python>=4.2",
        "opencv-python>=4.2",
        "pandas",  # do not change
        "Pillow>=5.4.1",
        "psutil",
        "PyQt5>=5.9.2,<=5.12",
        "python-dateutil>=2.8.0",
        "QtPy>=1.6.0",
        "qtap",
        "scikit-image~=0.15.0",  # do not change
        "scikit-learn~=0.23.1",  # tslearn 0.4 has issues with newer sklearn
        "scipy>=1.2.1",  # do not change
        "seaborn==0.9.0",  # do not change
        "spyder==3.3.3",  # do not change
        "tifffile",  # do not change
        "tqdm>=4.37.0",
        "PeakUtils",  # caiman requirement
        "tables>=3.6.1",
        "joblib>=0.15.1",
        #"tslearn~=0.2.2",  # had to remove because of windows
        "graphviz>=0.13",
        "numba>=0.50.0",
        #"bottleneck==1.2.1",  # do not change
        "holoviews",
        "jupyter",
        "pynwb",  # caiman requirement
        "pyqtgraph",
        #"tensorflow~=1.15.0",  # do not change, caiman requirement
        "tk",  # do not change
        "tcl",
        "bokeh<2.5.0,>=2.4.0",
        "nuset-lib",
        "sk-video",
        "click",
        "xmltodict"  # for OME xml
    ]

classifiers = \
    [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Intended Audience :: Science/Research"
    ]

with open("readme-pypi.md", 'r') as fh:
    long_description = fh.read()


with open(Path(__file__).parent.joinpath('mesmerize').joinpath('VERSION'), 'r') as vf:
    vesion_str = vf.read().split('\n')[0]


setup(
    name='mesmerize',
    version=vesion_str,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    entry_points=entry_points,
    url='https://github.com/kushalkolar/MESmerize',
    license='GNU General Public License v3.0',
    author='Kushal Kolar, Daniel Dondorp',
    author_email='kushalkolar@gmail.com',
    description='Calcium imaging analysis platform',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=classifiers,
    python_requires='>=3.6',
    install_requires=install_requires
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtWidgets
from ..core.common import ViewerUtils
from ...common.qdialogs import *
from ...common.configuration import get_sys_config
from qtap import Function
from caiman.source_extraction.cnmf.cnmf import CNMF, Estimates, CNMFParams
from .roi_manager_modules.managers import ManagerCNMFROI, ManagerVolCNMF, ManagerVolMultiCNMFROI
import numpy as np
from tqdm import tqdm
import logging


logger = logging.getLogger()


# just need the function signature
def detrend_df_f(quantileMin: float = 8,
                 frames_window: int = 500,
                 flag_auto: bool = True,
                 use_fast: bool = False,
                 use_residuals: bool = True,
                 detrend_only: bool = False):
    pass


# adapted from caiman code for reloading CNMF obj from hdf5 file
def get_CNMF_obj(d: dict) -> CNMF:
    new_obj = CNMF(get_sys_config()['_MESMERIZE_N_THREADS'])
    for key, val in d.items():
        if key == 'params':
            prms = CNMFParams()
            for subdict in val.keys():
                prms.set(subdict, val[subdict])
            setattr(new_obj, key, prms)
        elif key == 'estimates':
            estims = Estimates()
            for kk, vv in val.items():
                if kk == 'discarded_components':
                    if vv is not None:
                        discarded_components = Estimates()
                        for kk__, vv__ in vv.items():
                            setattr(discarded_components, kk__, vv__)
                        setattr(estims, kk, discarded_components)
                else:
                    setattr(estims, kk, vv)

            setattr(new_obj, key, estims)
        else:
            setattr(new_obj, key, val)

    return new_obj


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        QtWidgets.QDockWidget.__init__(self, parent)
        self.setFloating(True)
        self.setWindowTitle('Caiman detrend ΔF/F')

        self.vi = ViewerUtils(viewer_reference)

        opts = \
            {
                'quantileMin':
                    {
                        'minmax': (0, 100)
                    },
                'frames_window':
                    {
                        'minmax': (0, 1000),
                        'step': 10
                    }
            }

        self.function = Function(detrend_df_f, opts, parent=self)
        self.function.sig_set_clicked.connect(self.set_data)

        self.setWidget(self.function.widget)

    @present_exceptions(
        'Error getting dF/Fo data.',
        'The following error occurred when calculating dF/Fo values.'
    )
    def set_data(self, kwargs: dict):
        if not isinstance(
                self.vi.viewer.workEnv.roi_manager,
                (ManagerCNMFROI, ManagerVolCNMF, ManagerVolMultiCNMFROI)
        ):
            raise TypeError("This module is only for CNMF ROIs")

        roi_manager: (ManagerCNMFROI, ManagerCNMFROI, ManagerVolMultiCNMFROI) = \
            self.vi.viewer.workEnv.roi_manager

        self.vi.set_statusbar('Doing caiman detrend_df_f, please wait...')

        if isinstance(self.vi.viewer.workEnv.roi_manager, (ManagerCNMFROI, ManagerVolCNMF)):
            if not (roi_manager.idx_components == roi_manager.orig_idx_components).all():
                raise ValueError("You cannot use this module if you have deleted ROIs."
                                 "If you want to delete some ROIs, first use this "
                                 "module and then delete the ROIs.")

            self._set_data(roi_manager, kwargs)

        elif isinstance(self.vi.viewer.workEnv.roi_manager, ManagerVolMultiCNMFROI):
            for zlevel in range(len(roi_manager.idx_components)):
                if not (roi_manager.idx_components[zlevel] == roi_manager.orig_idx_components[zlevel]).all():
                    raise ValueError("You cannot use this module if you have deleted ROIs."
                                     "If you want to delete some ROIs, first use this "
                                     "module and then delete the ROIs.")

            self._set_data_multi2d(roi_manager, kwargs)

        history = self.vi.work_env.history_trace

        try:
            next(
                d for ix, d in enumerate(history) if 'caiman_detrend_df_f' in d
            )['caiman_detrend_df_f'] = kwargs
        except StopIteration:
            self.vi.work_env.history_trace.append({'caiman_detrend_df_f': kwargs})

        self.vi.set_statusbar('Finished caiman detrend_df_f!')

        self.vi.viewer.parent().get_module('roi_manager').ui.radioButton_dfof.click()

    def _set_data(
            self,
            roi_manager: Union[ManagerCNMFROI, ManagerVolCNMF],
            kwargs: dict
    ):
        cnmf = get_CNMF_obj(roi_manager.cnmf_data_dict)
        cnmf.estimates.detrend_df_f(**kwargs)

        roi_manager.cnm_dfof = cnmf.estimates.F_dff

        for roi, dfof in zip(
                roi_manager.roi_list,
                roi_manager.cnm_dfof
        ):
            xs = np.arange(len(dfof))
            roi.dfof_data = [xs, dfof]

    def _set_data_multi2d(
            self,
            roi_manager: ManagerVolMultiCNMFROI,
            kwargs: dict
    ):
        roi_manager.cnm_dfof.clear()

        for cnmf_data_dict in tqdm(roi_manager.cnmf_data_dicts, total=len(roi_manager.cnmf_data_dicts)):
            cnmf = get_CNMF_obj(cnmf_data_dict)
            cnmf.estimates.detrend_df_f(**kwargs)

            roi_manager.cnm_dfof.append(cnmf.estimates.F_dff)

        dfofs = np.vstack(roi_manager.cnm_dfof)

        logger.info(
            f"dfof shape:\t{dfofs.shape}\n"
            f"roi_list len:\t{len(roi_manager.roi_list)}"
        )

        for roi, f in zip(
                roi_manager.roi_list,
                dfofs
        ):
            xs = np.arange(len(f))
            roi.dfof_data = [xs, f]



import numpy as np
from scipy.cluster import hierarchy
from scipy.stats import wasserstein_distance
from ....analysis import Transmission
from .common import *
from typing import *
from ....analysis.math.emd import emd_1d
from sklearn.metrics import pairwise_distances
from scipy.spatial.distance import squareform
from ....common import get_sys_config


class Linkage(CtrlNode):
    """
    Basically scipy.cluster.hierarchy.linkage
    Compute a linkage matrix for Hierarchical clustering
    """

    nodeName = 'Linkage'
    uiTemplate = [('data_column',   'combo',    {'toolTip': 'Input column for clustering data, must form a 2D array'}),
                  ('method',        'combo',    {'items': ['complete', 'average', 'single']}),
                  ('metric',        'combo',    {'items': ['wasserstein', 'euclidean']}),
                  ('optimal_order', 'check',    {'checked': True}),
                  ('Apply',         'check',    {'checked': False, 'applyBox': True})]

    def processData(self, transmission: Transmission):
        self.t = transmission
        dcols = organize_dataframe_columns(self.t.df.columns.to_list())[0]
        self.ctrls['data_column'].setItems(dcols)

        if not self.ctrls['Apply'].isChecked():
            return

        self.data_column = self.ctrls['data_column'].currentText()

        self.data = np.vstack(self.t.df[self.data_column].values)

        method = self.ctrls['method'].currentText()
        metric = self.ctrls['metric'].currentText()
        optimal_ordering = self.ctrls['optimal_order'].isChecked()

        if metric == 'wasserstein':
            self.data += np.abs(self.data.min())
            distM = pairwise_distances(self.data, metric=emd_1d, n_jobs=get_sys_config()['_MESMERIZE_N_THREADS'])
            condensed = squareform(distM, checks=False)
            self.linkage = hierarchy.linkage(condensed, method=method, optimal_ordering=optimal_ordering)
        else:
            metric_ = metric
            self.linkage = hierarchy.linkage(self.data, method=method, metric=metric_, optimal_ordering=optimal_ordering)

        params = {'method': method, 'metric': metric, 'optimal_ordering': optimal_ordering}

        return {'linkage': self.linkage, 'params': params}


class FCluster(CtrlNode):
    """
    Basically scipy.cluster.hierarchy.fcluster.
    Form flat clusters from the hierarchical clustering defined by the given linkage matrix.
    """

    nodeName = 'FCluster'
    uiTemplate = [('threshold', 'doubleSpin', {'step': 0.001, 'min': 0.0, 'max': 9999.0}),
                  ('criterion', 'combo', {'items': ['distance', 'maxclust', 'inconsistent',
                                                    'monocrit', 'maxclust_monocrit']}),
                  ('depth', 'intSpin', {'min': 1, 'step': 1})
                  ]

    def __init__(self, name, **kwargs):
        CtrlNode.__init__(self, name, terminals={'Linkage': {'io': 'in'},
                                                 'Data':    {'io': 'in'},
                                                 'IncM':    {'io': 'in'},
                                                 'Monocrit':{'io': 'in'},

                                                 'Out': {'io': 'out'}}, **kwargs)

    def process(self, **kwargs):
        out = self.processData(**kwargs)
        return {'Out': out}

    def processData(self, **inputs):
        self.linkage = inputs['Linkage']

        if self.linkage is None:
            raise ValueError('Must input Linkage matrix')

        linkage_matrix = self.linkage['linkage']
        linkage_params = self.linkage['params']

        self.t = inputs['Data'].copy()

        if 'IncM' in inputs.keys():
            self.IncM = inputs['IncM']
        if 'Monocrit' in inputs.keys():
            self.Monocrit = inputs['Monocrit']

        self.threshold = self.ctrls['threshold'].value()
        self.criterion = self.ctrls['criterion'].currentText()

        self.depth = self.ctrls['depth'].value()

        R = None
        monocrit = None

        params = {'threshold':  self.threshold,
                  'criterion':  self.criterion,
                  'depth':      self.depth,
                  'linkage_matrix': linkage_matrix.tolist(),
                  'linkage_params': linkage_params}

        if self.criterion == 'distance':
            self.cluster_labels = hierarchy.fcluster(linkage_matrix, t=self.threshold, criterion='distance')
            self.t.df['FCLUSTER_LABELS'] = self.cluster_labels
            self.t.last_output = 'fcluster'
            self.t.history_trace.add_operation('all', operation='fcluster', parameters=params)
            return self.t

        elif self.criterion == 'maxclust':
            self.cluster_labels = hierarchy.fcluster(linkage_matrix, t=int(self.threshold), criterion='maxclust')
            self.t.df['FCLUSTER_LABELS'] = self.cluster_labels
            self.t.last_output = 'fcluster'
            self.t.history_trace.add_operation('all', operation='fcluster', parameters=params)
            return self.t

        else:
            params.update({'inc_M_levels': self.IncM['levels']})
            R = self.IncM['R']


            if self.criterion == 'monocrit':
                monocrit = self.Monocrit['MR']

            elif self.criterion == 'maxclust_monocrit':
                pass


class Inconsistent(CtrlNode):
    """Calculate inconsistency statistics on a linkage matrix. Returns inconsistency matrix"""
    nodeName = 'Inconsistent'
    uiTemplate = [('levels', 'intSpin', {'value': 2, 'min': 1, 'max': 99, 'step': 1})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'Linkage': {'io': 'in', 'multi': False}, 'IncM': {'io': 'out'}})

    def process(self, **kwargs) -> Dict[str, Dict[np.ndarray, int]]:
        Z = kwargs['Linkage']
        d = self.ctrls['levels']

        R = hierarchy.inconsistent(Z, d)

        return {'IncM': {'R': R, 'levels': d}}


class MaxInconsistent(CtrlNode):
    """Return the maximum inconsistency coefficient for each non-singleton cluster and its children."""
    nodeName = 'MaxInconsistent'
    uiTemplate = [('no params', 'label', {'text': 'no params'})]

    def __init__(self, name, **kwargs):
        CtrlNode.__init__(self, name, terminals={'Linkage': {'io': 'in', 'multi': False},
                                                 'IncM':    {'io': 'in', 'multi': False},
                                                 'MaxInc':     {'io': 'out'}}, **kwargs)

    def process(self, **kwargs):
        Z = kwargs['Linkage']
        IncM = kwargs['IncM']
        R = IncM['R']

        MI = hierarchy.maxinconsts(Z, R)

        return {'MaxInc': MI}


# class MaxDists(CtrlNode):
#     """Return the maximum distance between any non-singleton cluster."""
#     nodeName = 'MaxDists'
#     uiTemplate = [('no params', 'label', {'text', 'no params'})]
#
#     def __init__(self, name, **kwargs):
#         CtrlNode.__init__(self, name, terminals=)


class MaxIncStat(CtrlNode):
    """Return the maximum statistic for each non-singleton cluster and its children."""
    nodeName = 'MaxIncStat'
    uiTemplate = [('Stat', 'combo', {'items': ['mean', 'stdev', 'num_links', 'inc_coef']})]

    def __init__(self, name, **kwargs):
        CtrlNode.__init__(self, name, terminals={'Linkage', {'io': 'in', 'multi': False},
                                                 'IncM',    {'io': 'in', 'multi': False},
                                                 'MaxStat', {'io': 'out'}}, **kwargs)

    def process(self, **kwargs) -> Dict[str, Dict[np.ndarray, str]]:
        Z = kwargs['Linkage']
        IncM = kwargs['IncM']
        R = IncM['R']

        stat = self.ctrls['Stat'].currentText()

        if stat == 'mean':
            i = 0
        elif stat == 'stdev':
            i = 1
        elif stat == 'num_links':
            i = 2
        elif stat == 'inc_coef':
            i = 3

        MR = hierarchy.maxRstat(Z, R, i)

        return {'MaxStat': {'MR': MR, 'stat': stat}}



from PyQt5 import QtCore, QtWidgets


class ListWidget(QtWidgets.QListWidget):
    """QListWidget with a signal blocker for set_items"""

    def blockSignals(func, *args):
        def fn(self, *args, **kwds):
            blocked = self.signalsBlocked()
            self.blockSignals(True)
            try:
                ret = func(self, *args, **kwds)
            finally:
                self.blockSignals(blocked)

            return ret

        return fn

    @blockSignals
    def setItems(self, items: list):
        previous_items = [self.item(ix).text() for ix in range(self.count())]
        if set(items) == previous_items:
            return

        selected_items = [item.text() for item in self.selectedItems()]

        self.clear()
        self.addItems(items)

        for selected_item in selected_items:
            item = self.findItems(selected_item, QtCore.Qt.MatchExactly)
            if len(item) == 0:
                continue
            item = item[0]
            item.setSelected(True)

    def getSelectedItems(self) -> list:
        return [item.text() for item in self.selectedItems()]

    def setSelectedItems(self, items: list):
        for item in items:
            item = self.findItems(item, QtCore.Qt.MatchExactly)
            if len(item) == 0:
                continue
            item = item[0]
            item.setSelected(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 13:58:20 2017

@author: kushal, adapted from Daniel Dondorp

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import numpy as np
import traceback
#from common.misc_functions import *
import scipy.io as spio


class MatlabFuncs:
    @staticmethod
    def loadmat(filename: str) -> dict:
        data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
        return MatlabFuncs._check_keys(data)

    @staticmethod
    def _check_keys(d: dict) -> dict:
        for key in d:
            if isinstance(d[key], spio.matlab.mio5_params.mat_struct):
                d[key] = MatlabFuncs._todict(d[key])
        return d

    @staticmethod
    def _todict(matobj: spio.matlab.mio5_params.mat_struct) -> dict:
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                d[strg] = MatlabFuncs._todict(elem)
            else:
                d[strg] = elem
        return d


class MES:
    """
    Handles of opening .mes files and organizing the images and meta data.
    The load_img() method returns a 3D array (dims are [time, cols, rows])
    of the image sequence and its associated meta data.

    Usage:
    Create a MES instance by passing the path of your mes file, example:

    mesfile = MES('/path/to/mesfile/experiment_Feb_31.mes')

    Call the get_image_references() method to get a list of references for images that can be loaded.

    To load an image that is available in the instance, just pass one of the references from get_image_references()
    to the load_img method:

        img_array, meta_dict = mesfile.load_img('IF0001_0001')

    """

    def __init__(self, filename: str):
        # Open the weird matlab type objects and organize the images & meta data
        """
        :param filename: full path of a single .mes file
        """
        self.main_dict = MatlabFuncs.loadmat(filename)

        self.main_dict_keys = [x for x in self.main_dict.keys()]
        self.main_dict_keys.sort()
        self._images = [x for x in self.main_dict_keys if "I" in x]

        self.image_descriptions = {}

        self.voltages_lists_dict = {}
        self.errors = []

        for image in self._images:
            try:
                meta = self.main_dict["D" + image[1:6]].tolist()
                meta = self._todict(meta[int(image[-1]) - 1])
            except Exception:
                self.errors.append('Error opening an image, There as an error when opening the following image: ' + str(
                    "D" + image[1:6]) + '\n' + traceback.format_exc())

            # If auxiliary voltage information (which for example contains information
            # about stimulus timings & can be mapped to stimulus definitions).
            for prefix in ['AUXi', 'AUXo', 'Sh', 'Stim']:
                for n in range(0, 9):
                    channel = prefix + str(n)
                    if channel in meta.keys():
                        try:
                            if meta[channel]['y'].ndim == 2:
                                if channel in self.voltages_lists_dict.keys():
                                    self.voltages_lists_dict[channel] += list(set(meta[channel]['y'][1]))
                                else:
                                    self.voltages_lists_dict[channel] = list(set(meta[channel]['y'][1]))
                        except (KeyError, IndexError) as e:
                            self.errors.append(str(e) + ': ' + meta['IMAGE'] + ' Does not have: ' + channel)

            for channel in ['PMT_EN', 'Trig', 'PMTenUG', 'PMTenUR', 'PMTenUB']:
                if channel in meta.keys():
                    try:
                        if channel in self.voltages_lists_dict.keys():
                            self.voltages_lists_dict[channel] += list(set(meta[channel]['y'][1]))
                        else:
                            self.voltages_lists_dict[channel] = list(set(meta[channel]['y'][1]))
                    except (KeyError, IndexError) as e:
                        self.errors.append(str(e) + ': ' + meta['IMAGE'] + ' Does not have: ' + channel)
            try:
                comment = meta["Comment"]

                if type(comment) != str:
                    comment = "No description"
                self.image_descriptions[image] = comment
            except KeyError as e:
                self.errors.append(str(e) + ': ' + '"error for: "' + image)

            for channel in self.voltages_lists_dict.keys():
                self.voltages_lists_dict[channel] = list(set(self.voltages_lists_dict[channel]))

    def _loadmat(self, filename):
        data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
        return self._check_keys(data)

    def _check_keys(self, dict):
        for key in dict:
            if isinstance(dict[key], spio.matlab.mio5_params.mat_struct):
                dict[key] = self._todict(dict[key])
        return dict

    def _todict(self, matobj):
        dict = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                dict[strg] = self._todict(elem)
            else:
                dict[strg] = elem
        return dict

    def get_image_references(self) -> list:
        """Get a list of all image references available in the instance"""
        return self._images

    # Returns image as ImgData class object.
    def load_img(self, img_reference: str) -> (np.ndarray, dict):
        """
        :param img_reference: The image reference, usually something like IFxxxx_xxxx or Ifxxxx_xxxx
        :return: (image sequence array, meta data dict)
        """

        if img_reference not in self._images:
            raise KeyError('fImage reference: {img_reference} not found.\nCall get_image_references for a '
                           'list of all available references to images that can be loaded.')

        meta = self.main_dict["D" + img_reference[1:6]].tolist()
        meta = self._todict(meta[0])
        #        except KeyError:
        #            return False, KeyError

        if len(meta["FoldedFrameInfo"]) > 0:
            start = meta["FoldedFrameInfo"]["firstFramePos"]
            stop = meta["TransversePixNum"]
            # print("Images starting at: ",start)
            # print("Frame width = ",stop)
            im = self.main_dict[img_reference]
            # Trim the 2D array to start at where img acquisition actually begins
            im = im[:, start:]
            # Figure out where the 2D array stops at end of acquisition
            im = im[:, :(im.shape[1] - (im.shape[1] % stop))]
            # Divide the single large 2D array into individual arrays which each
            # represent a single frame
            image_sequence = np.hsplit(im, im.shape[1] / stop)
            # Combine all the frames into one 3D array, 2D + time
            seq = np.dstack(image_sequence)

            return seq, meta
        else:
            raise ValueError("'" + str(img_reference) + "' does not have the required "
                                                        "'FoldedFrameInfo' meta-data which is "
                                                        "required for the construction of an image "
                                                        "sequence from mesfile data")


# For testing
#if __name__ == '__main__':
# y = imdata.meta['AUXo3']['y']



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './ui_files/column_pytemplate.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_column_template(object):
    def setupUi(self, column_template):
        column_template.setObjectName("column_template")
        column_template.resize(200, 402)
        column_template.setMinimumSize(QtCore.QSize(200, 0))
        column_template.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.gridLayout = QtWidgets.QGridLayout(column_template)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.labelColumnName = QtWidgets.QLabel(column_template)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelColumnName.sizePolicy().hasHeightForWidth())
        self.labelColumnName.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.labelColumnName.setFont(font)
        self.labelColumnName.setText("")
        self.labelColumnName.setObjectName("labelColumnName")
        self.verticalLayout.addWidget(self.labelColumnName)
        self.listWidget = QtWidgets.QListWidget(column_template)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout.addWidget(self.listWidget)
        self.lineEdit = QtWidgets.QLineEdit(column_template)
        self.lineEdit.setObjectName("lineEdit")
        self.verticalLayout.addWidget(self.lineEdit)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 3)
        spacerItem = QtWidgets.QSpacerItem(67, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.btnReset = QtWidgets.QPushButton(column_template)
        self.btnReset.setMinimumSize(QtCore.QSize(50, 26))
        self.btnReset.setMaximumSize(QtCore.QSize(50, 26))
        self.btnReset.setObjectName("btnReset")
        self.gridLayout.addWidget(self.btnReset, 1, 1, 1, 1)
        self.btnApply = QtWidgets.QPushButton(column_template)
        self.btnApply.setMinimumSize(QtCore.QSize(50, 26))
        self.btnApply.setMaximumSize(QtCore.QSize(50, 26))
        self.btnApply.setObjectName("btnApply")
        self.gridLayout.addWidget(self.btnApply, 1, 2, 1, 1)

        self.retranslateUi(column_template)
        QtCore.QMetaObject.connectSlotsByName(column_template)

    def retranslateUi(self, column_template):
        _translate = QtCore.QCoreApplication.translate
        column_template.setWindowTitle(_translate("column_template", "Form"))
        self.lineEdit.setPlaceholderText(_translate("column_template", "Filter"))
        self.btnReset.setText(_translate("column_template", "Reset"))
        self.btnApply.setText(_translate("column_template", "Apply"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pca_window.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PCA_Window(object):
    def setupUi(self, PCA_Window):
        PCA_Window.setObjectName("PCA_Window")
        PCA_Window.setEnabled(True)
        PCA_Window.resize(994, 733)
        self.centralwidget = QtWidgets.QWidget(PCA_Window)
        self.centralwidget.setEnabled(True)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_2.addWidget(self.label_4)
        self.labelFreq = QtWidgets.QLabel(self.centralwidget)
        self.labelFreq.setObjectName("labelFreq")
        self.horizontalLayout_2.addWidget(self.labelFreq)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.sliderMaxFreqIndex = QtWidgets.QSlider(self.centralwidget)
        self.sliderMaxFreqIndex.setMinimum(1)
        self.sliderMaxFreqIndex.setMaximum(1000)
        self.sliderMaxFreqIndex.setProperty("value", 400)
        self.sliderMaxFreqIndex.setOrientation(QtCore.Qt.Horizontal)
        self.sliderMaxFreqIndex.setObjectName("sliderMaxFreqIndex")
        self.verticalLayout_4.addWidget(self.sliderMaxFreqIndex)
        self.gridLayout.addLayout(self.verticalLayout_4, 0, 0, 1, 4)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_6 = QtWidgets.QLabel(self.centralwidget)
        self.label_6.setMaximumSize(QtCore.QSize(100, 16777215))
        self.label_6.setObjectName("label_6")
        self.verticalLayout_6.addWidget(self.label_6)
        self.comboBoxLog = QtWidgets.QComboBox(self.centralwidget)
        self.comboBoxLog.setMaximumSize(QtCore.QSize(100, 16777215))
        self.comboBoxLog.setObjectName("comboBoxLog")
        self.comboBoxLog.addItem("")
        self.comboBoxLog.addItem("")
        self.comboBoxLog.addItem("")
        self.verticalLayout_6.addWidget(self.comboBoxLog)
        self.gridLayout.addLayout(self.verticalLayout_6, 1, 1, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.spinBoxTrimStart = QtWidgets.QSpinBox(self.centralwidget)
        self.spinBoxTrimStart.setProperty("value", 1)
        self.spinBoxTrimStart.setObjectName("spinBoxTrimStart")
        self.verticalLayout.addWidget(self.spinBoxTrimStart)
        self.gridLayout.addLayout(self.verticalLayout, 1, 0, 1, 1)
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.labelDataColumn = QtWidgets.QLabel(self.centralwidget)
        self.labelDataColumn.setObjectName("labelDataColumn")
        self.verticalLayout_7.addWidget(self.labelDataColumn)
        self.comboBoxDataColumn = QtWidgets.QComboBox(self.centralwidget)
        self.comboBoxDataColumn.setObjectName("comboBoxDataColumn")
        self.verticalLayout_7.addWidget(self.comboBoxDataColumn)
        self.gridLayout.addLayout(self.verticalLayout_7, 1, 2, 1, 1)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.labelTargetColumn = QtWidgets.QLabel(self.centralwidget)
        self.labelTargetColumn.setObjectName("labelTargetColumn")
        self.verticalLayout_5.addWidget(self.labelTargetColumn)
        self.comboBoxTarget = QtWidgets.QComboBox(self.centralwidget)
        self.comboBoxTarget.setObjectName("comboBoxTarget")
        self.verticalLayout_5.addWidget(self.comboBoxTarget)
        self.gridLayout.addLayout(self.verticalLayout_5, 1, 3, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(754, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 1, 6, 1, 1)
        self.checkBoxLiveUpdate = QtWidgets.QCheckBox(self.centralwidget)
        self.checkBoxLiveUpdate.setObjectName("checkBoxLiveUpdate")
        self.gridLayout.addWidget(self.checkBoxLiveUpdate, 1, 4, 1, 1)
        self.btnPlot = QtWidgets.QPushButton(self.centralwidget)
        self.btnPlot.setObjectName("btnPlot")
        self.gridLayout.addWidget(self.btnPlot, 1, 5, 1, 1)
        PCA_Window.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(PCA_Window)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 994, 22))
        self.menubar.setObjectName("menubar")
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        PCA_Window.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(PCA_Window)
        self.statusbar.setObjectName("statusbar")
        PCA_Window.setStatusBar(self.statusbar)
        self.dockConsole = QtWidgets.QDockWidget(PCA_Window)
        self.dockConsole.setFeatures(QtWidgets.QDockWidget.DockWidgetFloatable|QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockConsole.setObjectName("dockConsole")
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.dockConsole.setWidget(self.dockWidgetContents)
        PCA_Window.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.dockConsole)
        self.actionConsole = QtWidgets.QAction(PCA_Window)
        self.actionConsole.setCheckable(True)
        self.actionConsole.setObjectName("actionConsole")
        self.menuView.addAction(self.actionConsole)
        self.menubar.addAction(self.menuView.menuAction())

        self.retranslateUi(PCA_Window)
        self.sliderMaxFreqIndex.valueChanged['int'].connect(self.labelFreq.setNum)
        self.actionConsole.toggled['bool'].connect(self.dockConsole.setVisible)
        QtCore.QMetaObject.connectSlotsByName(PCA_Window)

    def retranslateUi(self, PCA_Window):
        _translate = QtCore.QCoreApplication.translate
        PCA_Window.setWindowTitle(_translate("PCA_Window", "MainWindow"))
        self.label_4.setText(_translate("PCA_Window", "Max included frequency index:"))
        self.labelFreq.setText(_translate("PCA_Window", "400"))
        self.label_6.setText(_translate("PCA_Window", "Apply log"))
        self.comboBoxLog.setItemText(0, _translate("PCA_Window", "log2"))
        self.comboBoxLog.setItemText(1, _translate("PCA_Window", "log10"))
        self.comboBoxLog.setItemText(2, _translate("PCA_Window", "None"))
        self.label.setText(_translate("PCA_Window", "Trim start index upto"))
        self.labelDataColumn.setText(_translate("PCA_Window", "Data column"))
        self.labelTargetColumn.setText(_translate("PCA_Window", "Target column"))
        self.checkBoxLiveUpdate.setText(_translate("PCA_Window", "Live Update"))
        self.btnPlot.setText(_translate("PCA_Window", "Plot"))
        self.menuView.setTitle(_translate("PCA_Window", "View"))
        self.dockConsole.setWindowTitle(_translate("PCA_Window", "Console"))
        self.actionConsole.setText(_translate("PCA_Window", "Console"))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtGui, QtWidgets
from .pytemplates.script_editor_pytemplate import Ui_ViewerScriptEditor
from .script_editor_modules.syntax_editor import Highlighter
import os
from ...common.qdialogs import *
from ...common.configuration import HOME


class ModuleGUI(QtWidgets.QMainWindow):
    def __init__(self, parent, viewer_reference):
        QtWidgets.QMainWindow.__init__(self, parent=parent)
        self.parent = parent
        self.console = parent.console
        self.ui = Ui_ViewerScriptEditor()
        self.ui.setupUi(self)
        self.open_temp_script()

        self.connect_actions()

        # self.ui.pushButtonRun.clicked.connect(self.run_in_console)
        self.ui.actionRun.triggered.connect(self.run_in_console)
        self.setWindowTitle('Viewer - Script editor')

    def connect_actions(self):
        self.ui.actionNew.triggered.connect(self.new_file)
        self.ui.actionOpen.triggered.connect(self.open_file)
        self.ui.actionSave.triggered.connect(self.save_file)

    def run_in_console(self):
        text = self.ui.tabWidget.currentWidget().toPlainText()
        self.console.runCmd(text)

    def create_editor(self):
        font = QtGui.QFont()
        font.setFamily('Courier')
        font.setFixedPitch(True)
        font.setPointSize(10)

        editor = QtWidgets.QTextEdit()
        editor.setFont(font)

        return editor

    def add_tab(self, file_path) -> QtWidgets.QTextEdit:
        editor = self.create_editor()
        self.ui.tabWidget.addTab(editor, file_path)
        self.ui.tabWidget.setCurrentIndex(self.ui.tabWidget.count() - 1)
        if not os.path.isfile(file_path):
            open(file_path, 'w').close()
            text = ''
        elif os.path.isfile(file_path):
            with open(file_path, 'r') as f:
                text = f.read()
        else:
            QtWidgets.QMessageBox.warning(self, 'Invalid file', 'The file path is invalid, maybe it is a directory')
        self.ui.tabWidget.currentWidget().highlighter = Highlighter(editor.document())
        self.ui.tabWidget.currentWidget().setPlainText(text)
        return self.ui.tabWidget.currentWidget()

    def open_temp_script(self):
        home_dir = os.environ['HOME']
        tmp_script = home_dir + '/mtemp.py'
        if not os.path.isfile(tmp_script):
            open(tmp_script, 'w').close()
        self.add_tab(tmp_script)

    def save_file(self):
        path = self.ui.tabWidget.tabText(self.ui.tabWidget.currentIndex())
        text = self.ui.tabWidget.currentWidget().toPlainText()
        with open(path, 'w') as f:
            f.write(text)

    def new_file(self):
        path = QtWidgets.QFileDialog.getSaveFileName(self, 'Create a new file', os.environ[HOME], '(*.py)')
        if path == '':
            return
        if path[0].endswith('.py'):
            path = path[0]
        else:
            path = path[0] + '.py'

        self.add_tab(path)

    def open_file(self):
        path = QtWidgets.QFileDialog.getOpenFileName(self, 'Open file', os.environ[HOME], '(*.py)')
        if path == '':
            return
        self.add_tab(path[0])

if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    w = ModuleGUI()
    w.show()
    app.exec_()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './control_widget.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DockWidgetControls(object):
    def setupUi(self, DockWidgetControls):
        DockWidgetControls.setObjectName("DockWidgetControls")
        DockWidgetControls.resize(307, 801)
        DockWidgetControls.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.groupBox = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.radioButtonDataColumn = QtWidgets.QRadioButton(self.groupBox)
        self.radioButtonDataColumn.setChecked(True)
        self.radioButtonDataColumn.setObjectName("radioButtonDataColumn")
        self.horizontalLayout.addWidget(self.radioButtonDataColumn)
        self.comboBoxDataColumn = QtWidgets.QComboBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxDataColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxDataColumn.setSizePolicy(sizePolicy)
        self.comboBoxDataColumn.setObjectName("comboBoxDataColumn")
        self.horizontalLayout.addWidget(self.comboBoxDataColumn)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.radioButtonXY = QtWidgets.QRadioButton(self.groupBox)
        self.radioButtonXY.setObjectName("radioButtonXY")
        self.verticalLayout.addWidget(self.radioButtonXY)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setObjectName("label")
        self.horizontalLayout_9.addWidget(self.label)
        self.comboBoxXColumn = QtWidgets.QComboBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxXColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxXColumn.setSizePolicy(sizePolicy)
        self.comboBoxXColumn.setObjectName("comboBoxXColumn")
        self.horizontalLayout_9.addWidget(self.comboBoxXColumn)
        self.verticalLayout.addLayout(self.horizontalLayout_9)
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.label_8 = QtWidgets.QLabel(self.groupBox)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_10.addWidget(self.label_8)
        self.comboBoxYColumn = QtWidgets.QComboBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxYColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxYColumn.setSizePolicy(sizePolicy)
        self.comboBoxYColumn.setObjectName("comboBoxYColumn")
        self.horizontalLayout_10.addWidget(self.comboBoxYColumn)
        self.verticalLayout.addLayout(self.horizontalLayout_10)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.checkBoxLogX = QtWidgets.QCheckBox(self.groupBox)
        self.checkBoxLogX.setObjectName("checkBoxLogX")
        self.horizontalLayout_12.addWidget(self.checkBoxLogX)
        self.checkBoxLogY = QtWidgets.QCheckBox(self.groupBox)
        self.checkBoxLogY.setObjectName("checkBoxLogY")
        self.horizontalLayout_12.addWidget(self.checkBoxLogY)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_12.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout_12)
        self.verticalLayout_2.addWidget(self.groupBox)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.comboBoxColorsColumn = QtWidgets.QComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxColorsColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxColorsColumn.setSizePolicy(sizePolicy)
        self.comboBoxColorsColumn.setObjectName("comboBoxColorsColumn")
        self.horizontalLayout_2.addWidget(self.comboBoxColorsColumn)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.label_7 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_2.addWidget(self.label_7)
        self.listWidgetColormap = ColormapListWidget(self.dockWidgetContents)
        self.listWidgetColormap.setObjectName("listWidgetColormap")
        self.verticalLayout_2.addWidget(self.listWidgetColormap)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.comboBoxShapesColumn = QtWidgets.QComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxShapesColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxShapesColumn.setSizePolicy(sizePolicy)
        self.comboBoxShapesColumn.setObjectName("comboBoxShapesColumn")
        self.horizontalLayout_3.addWidget(self.comboBoxShapesColumn)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_4 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_4.addWidget(self.label_4)
        self.comboBoxUUIDColumn = QtWidgets.QComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxUUIDColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxUUIDColumn.setSizePolicy(sizePolicy)
        self.comboBoxUUIDColumn.setObjectName("comboBoxUUIDColumn")
        self.horizontalLayout_4.addWidget(self.comboBoxUUIDColumn)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.label_9 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_11.addWidget(self.label_9)
        self.comboBoxDPTCurve = QtWidgets.QComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxDPTCurve.sizePolicy().hasHeightForWidth())
        self.comboBoxDPTCurve.setSizePolicy(sizePolicy)
        self.comboBoxDPTCurve.setObjectName("comboBoxDPTCurve")
        self.horizontalLayout_11.addWidget(self.comboBoxDPTCurve)
        self.verticalLayout_2.addLayout(self.horizontalLayout_11)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_5 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.spinBoxSpotSize = QtWidgets.QSpinBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.spinBoxSpotSize.sizePolicy().hasHeightForWidth())
        self.spinBoxSpotSize.setSizePolicy(sizePolicy)
        self.spinBoxSpotSize.setMinimum(1)
        self.spinBoxSpotSize.setMaximum(999)
        self.spinBoxSpotSize.setSingleStep(1)
        self.spinBoxSpotSize.setProperty("value", 10)
        self.spinBoxSpotSize.setObjectName("spinBoxSpotSize")
        self.horizontalLayout_5.addWidget(self.spinBoxSpotSize)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_6 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_6.addWidget(self.label_6)
        self.doubleSpinBoxAlpha = QtWidgets.QDoubleSpinBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.doubleSpinBoxAlpha.sizePolicy().hasHeightForWidth())
        self.doubleSpinBoxAlpha.setSizePolicy(sizePolicy)
        self.doubleSpinBoxAlpha.setMaximum(1.0)
        self.doubleSpinBoxAlpha.setSingleStep(0.05)
        self.doubleSpinBoxAlpha.setProperty("value", 1.0)
        self.doubleSpinBoxAlpha.setObjectName("doubleSpinBoxAlpha")
        self.horizontalLayout_6.addWidget(self.doubleSpinBoxAlpha)
        self.verticalLayout_2.addLayout(self.horizontalLayout_6)
        self.checkBoxLiveUpdateFromFlowchart = QtWidgets.QCheckBox(self.dockWidgetContents)
        self.checkBoxLiveUpdateFromFlowchart.setChecked(True)
        self.checkBoxLiveUpdateFromFlowchart.setObjectName("checkBoxLiveUpdateFromFlowchart")
        self.verticalLayout_2.addWidget(self.checkBoxLiveUpdateFromFlowchart)
        self.pushButtonUpdatePlot = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonUpdatePlot.setObjectName("pushButtonUpdatePlot")
        self.verticalLayout_2.addWidget(self.pushButtonUpdatePlot)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.pushButtonSave = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonSave.setObjectName("pushButtonSave")
        self.horizontalLayout_7.addWidget(self.pushButtonSave)
        self.pushButtonLoad = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonLoad.setObjectName("pushButtonLoad")
        self.horizontalLayout_7.addWidget(self.pushButtonLoad)
        self.verticalLayout_2.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.pushButton = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout_8.addWidget(self.pushButton)
        self.pushButtonExport = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonExport.setObjectName("pushButtonExport")
        self.horizontalLayout_8.addWidget(self.pushButtonExport)
        self.verticalLayout_2.addLayout(self.horizontalLayout_8)
        DockWidgetControls.setWidget(self.dockWidgetContents)

        self.retranslateUi(DockWidgetControls)
        QtCore.QMetaObject.connectSlotsByName(DockWidgetControls)
        DockWidgetControls.setTabOrder(self.comboBoxDataColumn, self.comboBoxColorsColumn)
        DockWidgetControls.setTabOrder(self.comboBoxColorsColumn, self.listWidgetColormap)
        DockWidgetControls.setTabOrder(self.listWidgetColormap, self.comboBoxShapesColumn)
        DockWidgetControls.setTabOrder(self.comboBoxShapesColumn, self.comboBoxUUIDColumn)
        DockWidgetControls.setTabOrder(self.comboBoxUUIDColumn, self.spinBoxSpotSize)
        DockWidgetControls.setTabOrder(self.spinBoxSpotSize, self.doubleSpinBoxAlpha)
        DockWidgetControls.setTabOrder(self.doubleSpinBoxAlpha, self.checkBoxLiveUpdateFromFlowchart)
        DockWidgetControls.setTabOrder(self.checkBoxLiveUpdateFromFlowchart, self.pushButtonUpdatePlot)
        DockWidgetControls.setTabOrder(self.pushButtonUpdatePlot, self.pushButtonSave)
        DockWidgetControls.setTabOrder(self.pushButtonSave, self.pushButtonLoad)
        DockWidgetControls.setTabOrder(self.pushButtonLoad, self.pushButton)
        DockWidgetControls.setTabOrder(self.pushButton, self.pushButtonExport)

    def retranslateUi(self, DockWidgetControls):
        _translate = QtCore.QCoreApplication.translate
        DockWidgetControls.setWindowTitle(_translate("DockWidgetControls", "Co&ntrols"))
        self.groupBox.setTitle(_translate("DockWidgetControls", "Data"))
        self.radioButtonDataColumn.setText(_translate("DockWidgetControls", "Data Colu&mn"))
        self.radioButtonXY.setText(_translate("DockWidgetControls", "Separate &X and Y"))
        self.label.setText(_translate("DockWidgetControls", "X:"))
        self.label_8.setText(_translate("DockWidgetControls", "Y:"))
        self.checkBoxLogX.setText(_translate("DockWidgetControls", "log x"))
        self.checkBoxLogY.setText(_translate("DockWidgetControls", "log y"))
        self.label_2.setText(_translate("DockWidgetControls", "Colors based on"))
        self.label_7.setText(_translate("DockWidgetControls", "Choose colormap:"))
        self.label_3.setText(_translate("DockWidgetControls", "Shapes based on"))
        self.label_4.setText(_translate("DockWidgetControls", "UUID column"))
        self.label_9.setText(_translate("DockWidgetControls", "DPT Curve column"))
        self.label_5.setText(_translate("DockWidgetControls", "Spot size"))
        self.label_6.setText(_translate("DockWidgetControls", "Alpha"))
        self.checkBoxLiveUpdateFromFlowchart.setText(_translate("DockWidgetControls", "Live update from flowchart"))
        self.pushButtonUpdatePlot.setText(_translate("DockWidgetControls", "Update Plot"))
        self.pushButtonSave.setText(_translate("DockWidgetControls", "Save"))
        self.pushButtonLoad.setText(_translate("DockWidgetControls", "Load"))
        self.pushButton.setText(_translate("DockWidgetControls", "Export to matplotlib"))
        self.pushButtonExport.setText(_translate("DockWidgetControls", "Export data"))

from ...utils import ColormapListWidget



# -*- coding: utf-8 -*-
from ...Qt import QtWidgets
# from . import functions
from .common import *
from ....plotting.widgets import CurvePlotWindow
# from analysis import pca_gui


class CurvePlotsNode(CtrlNode):
    """Curve plots"""
    nodeName = 'CurvePlots'
    uiTemplate = [('Apply', 'check', {'checked': False, 'applyBox': True}),
                  ('ShowGUI', 'button', {'text': 'OpenGUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
        self.plot_gui = None
        self.ctrls['ShowGUI'].clicked.connect(self._open_plot_gui)

    def process(self, **kwargs):
        if (self.ctrls['Apply'].isChecked() is False) or self.plot_gui is None:
            return

        transmissions = kwargs['In']

        if not len(transmissions) > 0:
            raise Exception('No incoming transmissions')

        transmissions_list = []

        for t in transmissions.items():
            t = t[1]
            if t is None:
                QtWidgets.QMessageBox.warning(None, 'None transmission', 'One of your transmissions is None')
                continue

            transmissions_list.append(t.copy())

        self.plot_gui.update_input_transmissions(transmissions_list)

    def _open_plot_gui(self):
        if self.plot_gui is None:
            self.plot_gui = CurvePlotWindow(parent=self.parent())
        self.plot_gui.show()

# class PCA(CtrlNode):
#     """PCA (Principal component analysis)"""
#     nodeName = 'PCA'
#     uiTemplate = [('Apply', 'check', {'checked': False, 'applyBox': True}),
#                   ('ShowGUI', 'button', {'text': 'OpenGUI'})]
#
#     def __init__(self, name):
#         CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}})
#         self.ctrls['ShowGUI'].clicked.connect(self._open_pca_gui)
#
#     def process(self, **kwargs):
#         if (self.ctrls['Apply'].isChecked() is False) or not hasattr(self, 'pca_gui'):
#             return
#
#         transmissions = kwargs['In']
#
#         if not len(transmissions) > 0:
#             raise Exception('No incoming transmissions')
#
#         transmissions_list = []
#
#         for t in transmissions.items():
#             t = t[1]
#             if t is None:
#                 QtWidgets.QMessageBox.warning(None, 'None transmission', 'One of your transmissions is None')
#                 continue
#
#             if not any('Power Spectral Density' in d for d in t.src):
#                 raise KeyError('Cannot perform PCA with incoming transmissions. '
#                                'You must pass through one of the following nodes before performing a PCA:\n'
#                                'Power Spectral Density')
#
#             transmissions_list.append(t.copy())
#
#         self.pca_gui.update_input(transmissions_list)
#
#     def _open_pca_gui(self):
#         if hasattr(self, 'pca_gui'):
#             self.pca_gui.show()
#             return
#
#         self.pca_gui = pca_gui.PCA_GUI()
#         self.pca_gui.show()
#         self.changed()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './ui_files/script_editor.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ViewerScriptEditor(object):
    def setupUi(self, ViewerScriptEditor):
        ViewerScriptEditor.setObjectName("ViewerScriptEditor")
        ViewerScriptEditor.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(ViewerScriptEditor)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.gridLayout.addWidget(self.tabWidget, 0, 0, 1, 2)
        ViewerScriptEditor.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(ViewerScriptEditor)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 20))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        ViewerScriptEditor.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(ViewerScriptEditor)
        self.statusbar.setObjectName("statusbar")
        ViewerScriptEditor.setStatusBar(self.statusbar)
        self.actionNew = QtWidgets.QAction(ViewerScriptEditor)
        self.actionNew.setObjectName("actionNew")
        self.actionOpen = QtWidgets.QAction(ViewerScriptEditor)
        self.actionOpen.setObjectName("actionOpen")
        self.actionSave = QtWidgets.QAction(ViewerScriptEditor)
        self.actionSave.setObjectName("actionSave")
        self.actionSave_as = QtWidgets.QAction(ViewerScriptEditor)
        self.actionSave_as.setObjectName("actionSave_as")
        self.actionRun = QtWidgets.QAction(ViewerScriptEditor)
        self.actionRun.setObjectName("actionRun")
        self.menuFile.addAction(self.actionNew)
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionSave_as)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionRun)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(ViewerScriptEditor)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(ViewerScriptEditor)

    def retranslateUi(self, ViewerScriptEditor):
        _translate = QtCore.QCoreApplication.translate
        ViewerScriptEditor.setWindowTitle(_translate("ViewerScriptEditor", "MainWindow"))
        self.menuFile.setTitle(_translate("ViewerScriptEditor", "&File"))
        self.actionNew.setText(_translate("ViewerScriptEditor", "&New"))
        self.actionNew.setShortcut(_translate("ViewerScriptEditor", "Ctrl+N"))
        self.actionOpen.setText(_translate("ViewerScriptEditor", "&Open"))
        self.actionOpen.setShortcut(_translate("ViewerScriptEditor", "Ctrl+O"))
        self.actionSave.setText(_translate("ViewerScriptEditor", "&Save"))
        self.actionSave.setShortcut(_translate("ViewerScriptEditor", "Ctrl+S"))
        self.actionSave_as.setText(_translate("ViewerScriptEditor", "S&ave as"))
        self.actionSave_as.setShortcut(_translate("ViewerScriptEditor", "Ctrl+Shift+S"))
        self.actionRun.setText(_translate("ViewerScriptEditor", "&Run"))
        self.actionRun.setShortcut(_translate("ViewerScriptEditor", "F5"))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 27 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

"""
import scipy.io as spio


def are_floats_equal(m: float, n: float, epsilon: float = 0.0001) -> bool:
    return abs((m + 0.0) - (n + 0.0)) < epsilon


class MatlabFuncs:
    @staticmethod
    def loadmat(filename: str) -> dict:
        data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
        return MatlabFuncs._check_keys(data)

    @staticmethod
    def _check_keys(d: dict) -> dict:
        for key in d:
            if isinstance(d[key], spio.matlab.mio5_params.mat_struct):
                d[key] = MatlabFuncs._todict(d[key])
        return d

    @staticmethod
    def _todict(matobj: spio.matlab.mio5_params.mat_struct) -> dict:
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                d[strg] = MatlabFuncs._todict(elem)
            else:
                d[strg] = elem
        return d



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...pyqtgraphCore import ScatterPlotItem, SpotItem, GraphicsLayoutWidget
from ...pyqtgraphCore.Qt import QtCore, QtGui, QtWidgets
# from ...pyqtgraphCore.functions import mkBrush
import pandas as pd
from uuid import UUID
# from collections.abc import Iterable
import numpy as np
from typing import *


class PgScatterPlot(QtCore.QObject):
    signal_spot_clicked = QtCore.pyqtSignal(UUID)  #: Emits the UUID of a spot when it is clicked

    def __init__(self, graphics_view: GraphicsLayoutWidget, parent=None):
        """
        :param graphics_view: This plot will instantiate within this GraphicsLayoutWidget
        """
        QtCore.QObject.__init__(self, parent)

        self.graphics_view = graphics_view
        self.current_datapoint = None
        self.lastClicked = []
        self.plots = self.graphics_view.addPlot(title='Scatter_Plot')
        self.plot = ScatterPlotItem(title='Scatter_Plot')
        self.plot.sigClicked.connect(self._clicked)

        self.plots.addItem(self.plot)

        self.legend = self.plots.addLegend()
        self.legend.setParentItem(self.plots)

        self.color_legend_items = []
        self.shape_legend_items = []

        # For creating legend
        self.pseudo_plots = []
        # self.legend.addItem(self.plot, 'Legend')
        
    def add_data(self, xs: np.ndarray, ys: np.ndarray, uuid_series: pd.Series,
                 color: Union[str, QtGui.QColor, QtGui.QBrush, List[Union[QtGui.QBrush, QtGui.QColor, str]]],
                 size: int = 10, **kwargs):
        """
        Add data to the plot

        :param xs:          array of x values, indices must correspond to the "ys" array
        :type xs:           np.ndarray

        :param ys:          array of y values, indices must correspond to the "xs" array
        :type ys:           np.ndarray

        :param uuid_series: series of UUID values.
                            Each SpotItem on the plot is tagged with these UUIDs, therefore
                            the indices must correspond to the "xs" and "ys" arrays.
        :type uuid_series:  pd.Series

        :param color:       Either a single color or list of colors that pqytgraph.fn.mkBrush() can accept
        :type color:        Union[str, QtGui.QColor, QtGui.QBrush, List[Union[QtGui.QBrush, QtGui.QColor, str]]]

        :param size:        spot size
        :type size:         int

        :param kwargs:      any additional kwargs that are passed to ScatterPlotItem.addPoints()
        """
        self.plot.addPoints(xs, ys, uuid=uuid_series, pen='k', brush=color, size=size, name='scatter', **kwargs)

    def _clicked(self, plot, points):
        """Called when the plot is clicked"""
        for i, p in enumerate(self.lastClicked):
            assert isinstance(p, SpotItem)
            p.setPen('k')#p._data['orig_pen'])
            p.setBrush(p._data['orig_brush'])
            # p.setSymbol(p._data['orig_symbol'])

        if len(points) == 1:
            p = points[0]
            assert isinstance(p, SpotItem)
            if isinstance(p.uuid, str):
                u = UUID(p.uuid)
            elif isinstance(p.uuid, UUID):
                u = p.uuid
            elif isinstance(p.uuid, int):
                u = p.uuid
            elif np.issubdtype(p.uuid, np.integer):
                u = p.uuid
            else:
                raise TypeError('spot uuid attribute must be either <uuid.UUID>, <str>, or <int>')

            self.signal_spot_clicked.emit(u)

        for p in points:
            assert isinstance(p, SpotItem)
            # p.setSymbol(p._data['orig_symbol'])
            p.setPen('w')

            p.setBrush('w')

        self.lastClicked = points

    def set_legend(self, colors: dict, shapes: dict = None):
        """
        Set the legend.

        :param colors: dict mapping of labels onto their corresponding colors {'label': QtGui.QColor}
        :type colors: Dict[str, Union[QtGui.QColor, QtGui.QBrush, str]]

        :param shapes: dict mapping of labels onto their corresponding shapes {'label': <shape as a str>}
        :type shapes: Dict[str, str]
        """
        self.clear_legend()

        for k in colors.keys():
            p = ScatterPlotItem()
            p.setData(x=[0], y=[0], brush=colors[k])
            self.color_legend_items.append(k)
            self.legend.addItem(p, k)
            self.pseudo_plots.append(p)

        if shapes is None:
            return

        for k in shapes.keys():
            p = ScatterPlotItem()
            p.setData(x=[0], y=[0], brush='w', symbol=shapes[k])
            self.shape_legend_items.append(k)
            self.legend.addItem(p, k)
            self.pseudo_plots.append(p)

    def clear_legend(self):
        """Clear the legend"""
        for i in range(len(self.pseudo_plots)):
            del self.pseudo_plots[0]

        for name in self.color_legend_items:
            self.legend.removeItem(name)
        self.color_legend_items.clear()

        for name in self.shape_legend_items:
            self.legend.removeItem(name)
        self.shape_legend_items.clear()

    def clear(self):
        """Clear the plot"""
        self.plot.clear()

    def export_plot(self, column: str, filetype: str = 'svg', title: str = None, error_bars: str = 'mean',
                    spots_color=None, spots_outline='black', background_color='black', axis_color='white'):
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 29 21:08:09 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

"""

from .mesfile import *
from .data_types import ImgData
from . import organize_metadata
import numpy as np
import pickle
import tifffile
import os
from ...common import get_sys_config, get_proj_config
from uuid import uuid4
from uuid import UUID as UUID_type
from typing import Optional, Tuple
import logging


logger = logging.getLogger()


class ViewerWorkEnv:
    def __init__(self, imgdata: ImgData = None, sample_id='', UUID=None, meta=None, stim_maps=None,
                 roi_manager=None, roi_states=None, comments='', origin_file='',
                 custom_cols = None, history_trace: list = None,
                 additional_data: dict = None,
                 misc: dict = None, **kwargs):
        """
        A class that encapsulates the main work environment objects (img sequence, ROIs, and ROI associated curves) of
        the viewer. Allows for a work environment to be easily spawned from different types of sources and allows for
        a work environment to be easily saved in different ways regardless of the type of original data source.

        :param roi_states:  roi states from ROI Manager module
        :param stim_maps:   {'units': str, 'dataframe': pd.DataFrame}
        :param history_trace: list of dicts containing a traceable history of what what done with the work environment,
                              such as params used from modules to process the data

        :type imgdata:      ImgData
        :type sample_id:    str
        :type UUID:         uuid.UUID
        :type meta:         dict
        :type stim_maps:    dict
        :type comments:     str
        :type cust_col_dict: dict
        :type roi_states:   dict
        :type history_trace: list

        """

        if imgdata is not None:
            self.isEmpty = False  #: Return True if the work environment is empty
            self.imgdata: ImgData = imgdata
            if isinstance(self.imgdata, ImgData):
                if meta is not None:
                    self.imgdata.meta = meta
        else:
            self.imgdata: ImgData = ImgData()  #: ImgData instance
            self.isEmpty = True
        self.meta = meta

        if history_trace is None:
            self.history_trace = []  #: history log
        else:
            self.history_trace = history_trace

        self.stim_maps = stim_maps  #: Stimulus maps

        if 'stimMaps' in kwargs.keys():
            self.stim_maps = kwargs['stimMaps']

        self.sample_id = sample_id  #: SampleID, if opened from a project Sample

        self.roi_manager = roi_manager  #: reference to the back-end ROI Manager that is currently in use

        self.changed_items = []
        self.roi_states = roi_states
        self.comments = comments
        self.origin_file = origin_file

        if custom_cols is None:
            self.custom_cols = {}

        if additional_data is not None:
            assert isinstance(additional_data, dict)
            self.additional_data = additional_data
        else:
            self.additional_data = {}

        if misc is not None:
            assert isinstance(misc, dict)
            self.misc = misc
        else:
            self.misc = {}

        self._UUID = UUID  #: UUID, if opened from a project Sample refers to the ImgUUID

        self._saved = True

    # def get_imgdata(self):
    #     return self.imgdata
    #
    # def get_seq(self):
    #     return self.imgdata.seq
    #
    # def get_meta(self):
    #     return self.meta
    #
    # def get_roi_manager(self):
    #     return self.roi_manager

    @property
    def UUID(self) -> UUID_type:
        if self._UUID is None:
            raise ValueError('UUID is not set. UUID can only be set at instantiation')
        if not isinstance(self._UUID, UUID_type):
            raise TypeError('UUID is not of type uuid.UUID. Something went wrong. You should not modify the UUID')
        return self._UUID

    def clear(self):
        """Cleanup of the work environment"""
        self.isEmpty = True
        # self.imgdata.clear()
        del self.imgdata
        self.imgdata = None
        if self.roi_manager is not None:
            self.roi_manager.clear()
            # self.roi_manager.parent.start_manual_mode()
        self.roi_states = None
        self.comments = ''
        self.origin_file = ''
        self._saved = True
        self.changed_items = []
        logger.info('Work environment dumped!')

    def restore_rois_from_states(self):
        if (self.roi_manager is not None) and (self.roi_states is not None):
            self.roi_manager.parent.set_all_from_states(self.roi_states)

    @classmethod
    def from_pickle(cls, pickle_file_path: str, tiff_path: str = None):
        """
        Get pickled image data from a pickle file & image sequence from a npz or tiff. Used after motion correction
        & to view a sample from a project DataFrame. Create ImgData class object (See MesmerizeCore.DataTypes) and
        return instance of the work environment.

        :param: pickle_file_path:   full path to the pickle containing image metadata, stim maps, and roi_states
        :param: tiff_path:          str of the full path to a tiff file containing the image sequence
        """

        if tiff_path is not None:
            # seq = tifffile.imread(tiff_path)
            tif = tifffile.TiffFile(tiff_path, is_nih=True)
            if len(tif.series) > 1:
                seq = tif.asarray(
                    key=range(0, len(tif.series)),
                    maxworkers=int(get_sys_config()['_MESMERIZE_N_THREADS'])
                )
            else:
                seq = tif.asarray(maxworkers=int(get_sys_config()['_MESMERIZE_N_THREADS']))
            if seq.ndim == 4:
                # tzxy to xytz
                seq = np.moveaxis(seq, (0, 1, 2, 3), (2, 3, 0, 1))
            else:
                seq = seq.T
        else:
            seq = None

        p = pickle.load(open(pickle_file_path, 'rb'))

        # compatibility for older pickle files from older mesmerize
        if 'imdata' in p.keys():
            try:
                sample_id = p['imdata']['sample_id']
            except KeyError:
                sample_id = p['imdata']['SampleID']

            imdata = ImgData(seq, p['imdata']['meta'])

            comments = p['imdata']['comments']

            roi_states = []
            if 'roi_states' in p['imdata']:
                for ID in range(0, len(p['imdata']['roi_states'])):
                    roi_states.append(p['imdata']['roi_states'][ID])

            return cls(imdata, roi_states=roi_states, comments=comments, sample_id=sample_id)
        else:
            # Use with output of new to_pickle() method
            img_data = ImgData(seq)
            return cls(img_data, **p)

    @property
    def saved(self):
        return bool(self._saved)

    @saved.setter
    def saved(self, state: bool):
        self._saved = state
        if state is True:
            self.changed_items = []

    @staticmethod
    def load_mesfile(path: str) -> MES:
        """
        Just passes the path of a .mes file to the constructor of class MES in MesmerizeCore.FileInput.
        Loads .mes file & constructs MES obj from which individual images & their respective metadata can be loaded
        to construct viewer work environments using the classmethod viewerWorkEnv.from_mesfile.

        :param path: full path to a single .mes file.
        """
        return MES(path)

    @staticmethod
    def _organize_meta(meta: dict, origin: str) -> dict:
        """
        Organize input meta data dict into a uniform structure
        :param meta:    meta data dict, origin from a json file for example
        :param origin:  name of the origin source of the meta data, such a program or microscope etc.
        :return:        dict organized with keys that are used throughout Mesmerize.
        """

        if origin == 'mes':
            fps = float(1000/meta['FoldedFrameInfo']['frameTimeLength'])

            date_meta = meta['MeasurementDate'].split('.')
            ymd = date_meta[0] + date_meta[1] + date_meta[2]
            hms_ = date_meta[3].split(':')
            hms = hms_[0].split(' ')[1] + hms_[1] + hms_[2][:2]
            date = ymd + '_' + hms

            vmin = meta['LUTstruct']['lower']
            vmax = meta['LUTstruct']['upper']

            meta_d = \
                {
                    'origin': origin,
                    'fps': fps,
                    'date': date,
                    'vmin': vmin,
                    'vmax': vmax,
                    'orig_meta': meta  # just the entire original meta data dict
                }

            return meta_d

        elif origin == 'AwesomeImager' or origin == 'DeepEyes':
            meta_d = \
                {
                    'origin': origin,
                    'version': meta['version'],
                    'fps': meta['framerate'],
                    'date': meta['date'] + '_' + meta['time'],
                    'vmin': meta['level_min'],
                    'vmax': meta['level_max'],
                    'orig_meta': meta  # just the entire original meta data dict
                }
            return meta_d

        else:
            required = ['fps', 'date']

            if not all(k in meta.keys() for k in required):
                raise KeyError(f'Meta data dict must contain all mandatory fields: {required}')

            meta_d = {'origin': origin,
                      'fps':    meta['fps'],
                      'date':   meta['date']}

            if 'orig_meta' in meta.keys():
                meta_d.update({'orig_meta': meta['orig_meta']})

            return meta_d

    @classmethod
    def from_mesfile(cls, mesfile_object: MES, img_reference: str):
        """
        Return instance of work environment with MesmerizeCore.ImgData class object using seq returned from
        MES.load_img from MesmerizeCore.FileInput module and any stimulus map that the user may have specified.

        :param mesfile_object: MES object, created from .mes file
        :param img_reference: image reference to load, see :meth:`mesmerize.viewer.core.mesfile.MES.get_image_references`
        """

        imgseq, raw_meta = mesfile_object.load_img(img_reference)

        fps = float(1000 / raw_meta['FoldedFrameInfo']['frameTimeLength'])

        date_meta = raw_meta['MeasurementDate'].split('.')
        ymd = date_meta[0] + date_meta[1] + date_meta[2]
        hms_ = date_meta[3].split(':')
        hms = hms_[0].split(' ')[1] + hms_[1] + hms_[2][:2]
        date = ymd + '_' + hms

        vmin = raw_meta['LUTstruct']['lower']
        vmax = raw_meta['LUTstruct']['upper']

        meta = \
            {
                'origin': 'mes',
                'fps': fps,
                'date': date,
                'vmin': vmin,
                'vmax': vmax,
                'orig_meta': raw_meta  # just the entire original meta data dict
            }

        imdata = ImgData(imgseq, meta)
        # imdata.stimMaps = (mesfileMaps, 'mesfile')

        return cls(imdata)

    @classmethod
    def from_tiff(
            cls, path: str,
            method: str,
            meta_path: Optional[str] = None,
            axes_order: Optional[str] = None,
            meta_format: Optional[str] = None,
    ):
        """
        Return instance of work environment with ImgData.seq set from the tiff file.

        :param path:        path to the tiff file
        :param method:      one of 'imread', 'asarray', or 'asarray-multi'. Refers to usage of either tifffile.imread
                            or tifffile.asarray. 'asarray-multi' will load multi-page tiff files.
        :param meta_path:   path to a file containing meta data
        :param meta_format: meta data format, must correspond to the name of a function in viewer.core.organize_meta
        :param axes_order:  Axes order as a 3 or 4 letter string for 2D or 3D data respectively.
                            Axes order is assumed to be "txy" or "tzxy" if not specified.
        """

        if method == 'imread':
            seq = tifffile.imread(path)

        elif method == 'asarray':
            tif = tifffile.TiffFile(path, is_nih=True)
            seq = tif.asarray(maxworkers=int(get_sys_config()['_MESMERIZE_N_THREADS']))

        elif method == 'asarray-multi':
            tif = tifffile.TiffFile(path, is_nih=True)

            seq = tif.asarray(key=range(0, len(tif.series)),
                              maxworkers=int(get_sys_config()['_MESMERIZE_N_THREADS']))
        else:
            raise ValueError("Must specify 'imread' or 'asarray' in method argument")

        if (meta_path is not None) and (meta_format is not None):
            try:
                meta = getattr(organize_metadata, meta_format)(meta_path)
            except Exception as e:
                raise TypeError(f"The meta data loader for <{meta_format}> was unable to "
                                f"load the following file:"
                                f"\n\n"
                                f"{meta_path}"
                                f"\n\n"
                                f"Make sure that you have chosen the correct `meta_format` "
                                f"for this file."
                                f"\n\n"
                                f"{traceback.format_exc()}")

            required = ['fps', 'date']

            if not all(k in meta.keys() for k in required):
                raise KeyError(f'Meta data dict must contain all mandatory fields: {required}')

        else:
            meta = None

        # Custom user mapping if specified
        if axes_order is not None:
            if not set(axes_order).issubset({'x', 'y', 't', 'z'}):
                raise ValueError('`axes_order` must only contain "x", "y", "t" and/or "z" characters')

            if len(axes_order) != seq.ndim:
                raise ValueError('number of dims specified by `axes_order` must match dims of image sequence')

            if seq.ndim == 4:
                new_axes = tuple(map({'x': 0, 'y': 1, 't': 2, 'z': 3}.get, axes_order))
                seq = np.moveaxis(seq, (0, 1, 2, 3), new_axes)
            else:
                new_axes = tuple(map({'x': 0, 'y': 1, 't': 2}.get, axes_order))
                seq = np.moveaxis(seq, (0, 1, 2), new_axes)

        # Default mapping
        else:
            # default axes remapping from tzxy
            if seq.ndim == 4:
                # tzxy to xytz
                seq = np.moveaxis(seq, (0, 1, 2, 3), (2, 3, 0, 1))
            else:
                # for 2D
                seq = seq.T

        imdata = ImgData(seq, meta)
        return cls(imdata)

    @classmethod
    def from_splits(cls):
        pass

    def _make_dict(self) -> dict:
        # Dict that's later used for pickling
        d = {'sample_id':   self.sample_id,
             'meta':        self.imgdata.meta,
             'stim_maps':   self.stim_maps,
             'comments':    self.comments,
             'history_trace': self.history_trace,
             'additional_data': self.additional_data
             }

        if self.roi_manager is not None:
            if not self.roi_manager.is_empty():
                rois = self._curate_roi_tags(
                    self.roi_manager.get_all_states()
                )

                d['roi_states'] = rois

        return d

    def _curate_roi_tags(self, rois):
        # I'm sorry
        if rois['roi_type'] == 'VolMultiCNMFROI':
            for z in range(len(rois['states'])):
                for ix in range(len(rois['states'][z])):
                    for roi_def in rois['states'][z][ix]['tags'].keys():
                        if rois['states'][z][ix]['tags'][roi_def] == '':
                            rois['states'][z][ix]['tags'][roi_def] = 'untagged'

            return rois

        else:
            for ix in range(len(rois['states'])):
                for roi_def in rois['states'][ix]['tags'].keys():
                    if rois['states'][ix]['tags'][roi_def] == '':
                        rois['states'][ix]['tags'][roi_def] = 'untagged'

            return rois

    def _prepare_export(
            self,
            dir_path: str,
            filename: Optional[str] = None,
            save_img_seq: bool = True,
            UUID: Optional[UUID_type] = None
        ) -> Tuple[str, dict]:

        if UUID is None:
            UUID = uuid4()

        if filename is None:
            filename = os.path.join(dir_path, f'{self.sample_id}-_-{UUID}')
        else:
            filename = os.path.join(dir_path, filename)

        work_env = self._make_dict()

        data = {**work_env, 'UUID': UUID}

        if save_img_seq:
            if self.imgdata.ndim == 3:
                tifffile.imsave(f'{filename}.tiff', self.imgdata.seq.T, bigtiff=True)

            elif self.imgdata.ndim == 4:
                tifffile.imsave(
                    f'{filename}.tiff',
                    np.moveaxis(
                        self.imgdata._seq,
                        (2, 3, 0, 1),  # from xytz
                        (0, 1, 2, 3)   # to   tzxy
                    ),
                    bigtiff=True
                )

        return (filename, data)

    def to_pickle(self, dir_path: str, filename: Optional[str] = None, save_img_seq=True, UUID=None) -> str:
        """
        Package the current work Env ImgData class object (See MesmerizeCore.DataTypes) and any paramteres such as
        for motion correction and package them into a pickle & image seq array. Used for batch motion correction and
        for saving current sample to the project. Image sequence is saved as a tiff and other information about the
        image is saved in a pickle.
        """
        filename, data = self._prepare_export(dir_path, filename, save_img_seq, UUID)

        pickle.dump(data, open(f'{filename}.pik', 'wb'), protocol=4)

        self.saved = True

        return filename

    def to_pandas(self, proj_path: str, modify_options: Optional[dict] = None) -> list:
        """
        Used for saving the work environment as a project Sample.

        :param      proj_path:      Root path of the current project
        :param      modify_options:
        :return:    list of dicts that each correspond to a single curve that can be appended
                    as rows to the project dataframe
        """
        if self.isEmpty:
            raise ValueError('Work environment is empty')

        if self.imgdata.meta['fps'] == 0 or self.imgdata.meta['fps'] == 0.0 or self.imgdata.meta['fps'] is None:
            raise AttributeError("Sampling rate of the image data is not set."
                                 "You must set `viewer.workEnv.imgdata.meta['fps'].\n"
                                 "If you are using the viewer console you can do:\n"
                                 "get_meta()['fps'] = <sampling rate as int or float>")

        save_img_seq = True

        # Path where image (as tiff file) and image metadata, roi_states, and stimulus maps (in a pickle) are stored
        imgdir = os.path.join(proj_path, 'images')  # + self.imgdata.SampleID + '_' + str(time.time())

        if (modify_options is dict) and (self._UUID is None):
            raise ValueError('Error overwriting Sample. Current Work Environment does not have a UUID.\n'
                             'Samples always have a UUID, something went wrong. Reload the sample from the project.')

        if self._UUID is None:
            UUID = uuid4()
        else:
            UUID = self.UUID

        if modify_options is not None:
            if modify_options['overwrite_img_seq']:
                save_img_seq = True
            else:
                save_img_seq = False

        img_path = self.to_pickle(imgdir, UUID=UUID, save_img_seq=save_img_seq)

        if save_img_seq:
            if self.imgdata.ndim == 3:
                max_proj = np.amax(self.imgdata.seq, axis=2)
                max_proj_path = img_path + '_max_proj.tiff'
                tifffile.imsave(max_proj_path, max_proj)

                std_proj = self.imgdata.seq.std(axis=2)
                std_proj_path = img_path + '_std_proj.tiff'
                tifffile.imsave(std_proj_path, std_proj)
            # projection for each zlevel
            elif self.imgdata.ndim == 4:
                for z in range(self.imgdata.z_max):
                    self.imgdata.set_zlevel(z)

                    max_proj = np.amax(self.imgdata.seq, axis=2)
                    max_proj_path = f'{img_path}_max_proj-{z}.tiff'
                    tifffile.imsave(max_proj_path, max_proj)

                    std_proj = self.imgdata.seq.std(axis=2)
                    std_proj_path = f'{img_path}_std_proj-{z}.tiff'
                    tifffile.imsave(std_proj_path, std_proj)

        # Since viewerWorkEnv.to_pickle sets the saved property to True, and we're not done saving the dict yet.
        self._saved = False

        # Create a dict that contains all stim definitions as keys that refer to a list of all the stims for that sample
        stimuli_unique_sets = {}
        # This list is just used for gathering all new stims to add to the config file. This is just used for
        # populating the comboBoxes in the stimMapWidget GUI so that the widget doesn't need to access the DataFrame
        # for this simple task.
        # new_stimuli = []
        if self.stim_maps is None:
            for stim_def in get_proj_config().options('STIM_DEFS'):
                stimuli_unique_sets[stim_def] = ['untagged']
        else:
            for stim_def in self.stim_maps.keys():
                stimuli = []

                if self.stim_maps[stim_def] is None:
                    stimuli_unique_sets[stim_def] = ['untagged']
                    continue

                stimuli += list(self.stim_maps[stim_def]['name'].values)

                stimuli_unique_sets[stim_def] = list(set(stimuli))

                # for stim in stimuli_unique_sets[stim_def]:
                #     if stim not in get_proj_config()['ALL_STIMS'].keys():
                        # new_stimuli.append(stim)

        # get_proj_config()['ALL_STIMS'] = {**get_proj_config()['ALL_STIMS'], **dict.fromkeys(new_stimuli)}
        # configuration.save_proj_config()

        if self.imgdata.meta is not None:
            try:
                date = str(self.imgdata.meta['date'])
            except KeyError:
                date = 'unknown'
        else:
            date = 'unknown'

        if self.comments is None or self.comments == '':
            comments = 'untagged'
        else:
            comments = self.comments

        dicts = []

        rois = self.roi_manager.get_all_states()

        if rois['roi_type'] != 'VolMultiCNMFROI':
            for ix in range(len(rois['states'])):

                for roi_def in rois['states'][ix]['tags'].keys():
                    if rois['states'][ix]['tags'][roi_def] == '':
                        rois['states'][ix]['tags'][roi_def] = 'untagged'

                roi_tags = rois['states'][ix]['tags']

                d = {'SampleID': self.sample_id,
                     'AnimalID': self.sample_id.split('-_-')[0],
                     'ImgUUID': str(UUID),
                     'ImgPath': os.path.relpath(f'{img_path}.tiff', proj_path),
                     'ImgInfoPath': os.path.relpath(f'{img_path}.pik', proj_path),
                     'ROI_State': rois['states'][ix],
                     'date': date,
                     'uuid_curve': str(uuid4()),
                     'comments': comments,
                     'misc': self.misc
                     }

                dicts.append({**d,
                              **self.custom_cols,
                              **stimuli_unique_sets,
                              **roi_tags})
        else:
            for z in range(len(rois['states'])):
                for ix in range(len(rois['states'][z])):

                    for roi_def in rois['states'][z][ix]['tags'].keys():
                        if rois['states'][z][ix]['tags'][roi_def] == '':
                            rois['states'][z][ix]['tags'][roi_def] = 'untagged'

                    roi_tags = rois['states'][z][ix]['tags']


                    d = {'SampleID': self.sample_id,
                         'AnimalID': self.sample_id.split('-_-')[0],
                         'ImgUUID': str(UUID),
                         'ImgPath': os.path.relpath(f'{img_path}.tiff', proj_path),
                         'ImgInfoPath': os.path.relpath(f'{img_path}.pik', proj_path),
                         'ROI_State': rois['states'][z][ix],
                         'date': date,
                         'uuid_curve': str(uuid4()),
                         'comments': comments,
                         'misc': self.misc
                         }

                    dicts.append({**d,
                                  **self.custom_cols,
                                  **stimuli_unique_sets,
                                  **roi_tags})

        self.saved = True
        return dicts



# -*- coding: utf-8 -*-
"""
Example beeswarm / bar chart
"""

from ...pyqtgraphCore import ScatterPlotItem, SpotItem, pseudoScatter, GraphicsLayoutWidget
from ...pyqtgraphCore.Qt import QtCore, QtGui, QtWidgets
import pandas as pd
from uuid import UUID


# TODO: ERROR BARS
class BeeswarmPlot(QtCore.QObject):
    signal_spot_clicked = QtCore.pyqtSignal(UUID)

    def __init__(self, graphics_view: GraphicsLayoutWidget, parent=None):
        # QtWidgets.QWidget.__init__(self)
        QtCore.QObject.__init__(self, parent)
        # layout = QtWidgets.QVBoxLayout(self)
        self.graphics_view = graphics_view
        self.plots = []
        self.scatter_plots = []
        # layout.addWidget(self.plot_widget)
        
        self.infinite_lines = []
        self.title = ''
        self.current_datapoint = None
        self.lastClicked = []
        self.pseudo_plots = []

        # self.legend = self.graphics_view
        # self.legend.setParentItem(self.plots)

        self.color_legend_items = []

    def add_plot(self, title: str):
        if len(self.scatter_plots) % 4 == 0:
            self.graphics_view.nextRow()
        plot = self.graphics_view.addPlot(title=title)
        scatter_plot = ScatterPlotItem(title=title)
        scatter_plot.sigClicked.connect(self._clicked)

        plot.addItem(scatter_plot)
        self.scatter_plots.append({'scatter_plot': scatter_plot, 'i': 0})
        self.plots.append(plot)

    def add_data_to_plot(self, ix: int, data_series: pd.Series, uuid_series: pd.Series, name, color):
        if ix > len(self.plots) - 1:
            raise IndexError('Plot index out of range.')

        not_nan = data_series.notna()
        yvals = data_series[not_nan].values
        xvals = pseudoScatter(yvals, spacing=0.025, bidir=True) * 0.4
        scatter_plot = self.scatter_plots[ix]['scatter_plot']
        self.scatter_plots[ix]['i'] += 1
        i = self.scatter_plots[ix]['i']
        scatter_plot.addPoints(x=xvals + i, y=yvals, uuid=uuid_series[not_nan], name=name, brush=color, pen='k', symbol='o', size=10)

    def _clicked(self, plot, points):
        for i, p in enumerate(self.lastClicked):
            assert isinstance(p, SpotItem)
            p.setPen('k')
            p.setBrush(p._data['orig_brush'])

        if len(points) == 1:
            p = points[0]
            assert isinstance(p, SpotItem)
            self.signal_spot_clicked.emit(p.uuid)

        for p in points:
            assert isinstance(p, SpotItem)
            p.setPen('w')
            p.setBrush('w')

        self.lastClicked = points

    def set_legend(self, colors: dict, shapes: dict = None):
        """
        :param colors: {'group_name': QtGui.QColor}
        :param shapes: {'group_name': <shape>}
        :return:
        """
        return
        self.clear_legend()

        for k in colors.keys():
            p = ScatterPlotItem()
            p.setData(x=[0], y=[0], brush=colors[k])
            self.color_legend_items.append(k)
            self.legend.addItem(p, k)
            self.pseudo_plots.append(p)

    def clear_legend(self):
        for i in range(len(self.pseudo_plots)):
            del self.pseudo_plots[0]

        for name in self.color_legend_items:
            self.legend.removeItem(name)
        self.color_legend_items.clear()
    
    def clear_plots(self):
        for plot in self.plots:
            self.graphics_view.removeItem(plot)
            plot.deleteLater()
        del self.plots
        self.plots = []
        del self.scatter_plots
        self.scatter_plots = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
import seaborn as sns
import matplotlib.pyplot as plt
from math import ceil


class ViolinsPlot(MatplotlibWidget):
    def __init__(self):
        MatplotlibWidget.__init__(self, size = (8,6), dpi=100)
#        self.fig.subplots_adjust(hspace=0.3, wspace=0.3)
        self.num_columns = 4
        self.axes = []

    def set(self, sub_dataframe, data_columns: list, x_column: str, sub_group_column: str = None, x_order: list = None, label_rotation: int = 45):
        self.axes = []
        self.fig.clear()

        num_rows = ceil(len(data_columns) / self.num_columns)

        for i, data_column in enumerate(data_columns):
            ax = self.fig.add_subplot(num_rows, self.num_columns, i + 1)
            self.axes.append(ax)
            sns.violinplot(x=x_column, y=data_column, hue=sub_group_column, order=x_order, palette='muted',
                                data=sub_dataframe[sub_dataframe[data_column].notna()], ax=self.axes[i])
            ax.set_xticklabels(x_order, rotation=label_rotation)
        self.fig.tight_layout()
        self.draw()

##Testing code
if __name__ == "__main__":
    import pandas as pd

    import sys
    from PyQt5 import QtWidgets
    iris = sns.load_dataset('iris')
    app = QtWidgets.QApplication([])
    v = ViolinsPlot()
    v.set(iris, [x for x in iris.columns if "species" not in x], "species", label_rotation=90)
    v.show()
    sys.exit(app.exec())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
import seaborn as sns
from pandas import DataFrame
import numpy as np


class TimeseriesPlot(MatplotlibWidget):
    def __init__(self):
        MatplotlibWidget.__init__(self)
        self.ax = self.fig.add_subplot(111)
        self.plot = None

        self.axs = []

    def set(self, data: np.ndarray, xticks: np.ndarray = None, **kwargs):
        """

        :param data:    2D Array of 1D curves
        :param xticks:
        :param kwargs:
        """
        self._set_lineplot(data, xticks=xticks, **kwargs)

    # def _set_tsplot(self):
    #     pass

    def _set_lineplot(self, data: np.ndarray, **kwargs):
        self.clear()
        df = DataFrame(data).melt()
        x = kwargs.pop('xticks')
        if x is not None:
            if x.size != data.shape[1]:
                raise ValueError('Length of xticks must match those of the input curves')

            xs = np.tile(x, data.shape[0])
            df.variable = xs

        self.plot = sns.lineplot(x='variable', y='value', data=df, ax=self.ax, **kwargs)
        self.draw()

    def set_single_line(self, y: np.ndarray, x: np.ndarray = None, **kwargs):
        self.clear()
        if x is None:
            x = np.linspace(0, len(y), len(y))

        self.plot = sns.lineplot(x=x, y=y, ax=self.ax, **kwargs)
        self.draw()

    def add_line(self, y: np.ndarray, x: np.ndarray = None, **kwargs):
        if x is None:
            x = np.linspace(0, len(y), len(y))

        self.axs.append(self.ax.twinx())
        sns.lineplot(x=x, y=y, ax=self.axs[-1], **kwargs)

        if 'color' in kwargs.keys():
            c = kwargs['color']
            self.axs[-1].tick_params(axis='y', colors=c)

    def clear(self):
        self.ax.cla()

    def clear_all(self):
        self.ax.cla()
        for ax in self.axs:
            ax.cla()
